// rate is BYTE/s

#include <stdio.h>
#include <limits.h>
#include "formula.h"
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>
#include <assert.h>
#include <float.h>

#include "smcc.h"

#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))

int hdr_smcc::offset_;
int hdr_smcc_ack::offset_;

nsaddr_t *bcl_addresses_smcc;
nsaddr_t *smcc_addresses_smcc;


static class SMCCHeaderClass : public PacketHeaderClass
{
	public:
		SMCCHeaderClass() : PacketHeaderClass ( "PacketHeader/SMCC",
				                                        sizeof ( hdr_smcc ) )
		{
			bind_offset ( &hdr_smcc::offset_ );
		}
} class_smcchdr;

static class SMCC_ACKHeaderClass : public PacketHeaderClass
{
	public:
		SMCC_ACKHeaderClass() : PacketHeaderClass ( "PacketHeader/SMCC_ACK",
				        sizeof ( hdr_smcc_ack ) )
		{
			bind_offset ( &hdr_smcc_ack::offset_ );
		}
} class_smcc_ackhdr;

static class SmccClass : public TclClass
{
	public:
		SmccClass() : TclClass ( "Agent/SMCC" ) {}
		TclObject* create ( int, const char*const* )
		{
			return ( new SmccAgent() );
		}
} class_smcc;


SmccAgent::SmccAgent() : Agent ( PT_SMCC )
{
	bind ( "packetSize_", &size_ );
	bind ( "fairSize_", &fairsize_ );
	bind ( "ndatapack_", &ndatapack_ );
	bind ( "InitRate_", &InitRate_ );
	bind ( "overhead_", &overhead_ );
	bind ( "ssmult_", &ssmult_ );
	bind ( "printStatus_", &printStatus_ );
	bind ( "bval_", &bval_ );
	bind ( "min_rate_", &min_rate_ );
	bind ( "t_factor_", &t_factor_ );
	bind ( "Initial_RTT_", &Initial_RTT_ );
	bind ( "sndr_id_", &sndr_id );
	bind ( "base_layer_rate", &base_layer_rate );

}

int SmccAgent::command ( int argc, const char*const* argv )
{
	if ( argc==2 )
	{
		if ( strcmp ( argv[1],"warmup" ) ==0 )
		{
			warmup();
			return TCL_OK;
		}
		if ( strcmp ( argv[1],"start" ) ==0 )
		{
			start();
			return TCL_OK;
		}
	}
	if ( argc == 3 )
	{
		if ( strcmp ( argv[1],"stop" ) ==0 )
		{
			stop ( atoi ( argv[2] ) );
			return TCL_OK;
		}
	}

	return ( Agent::command ( argc, argv ) );
}

void SmccAgent::start()
{
//	printf(" SmccAgent::start()\n");
	double now = Scheduler::instance().clock();
	printf ( " Base layer rate %f Mbps\n",base_layer_rate*8.0/1000000 );
	// send the first packet on first smcc layer
	sendpkt ( 0 );
	// ... at initial rate
	send_timer_[0]->resched ( size_/rate_[0] );
	send_timer_[0]->smcc_layer = 0;
	// ... and start timer so we can cut rate
	// in half if we do not get feedback
//	NoFeedbacktimer_[0].resched(DEC_NO_REPORT*size_/rate_[0]);
	NoFeedbacktimer_[0]->smcc_layer = 0;
	next_send_time = now + ( size_/rate_[0] );

}

void SmccAgent::warmup()
{
	Tcl & tcl = Tcl::instance();
	smcc_addresses_smcc = new nsaddr_t[Max_smcc_Layer];

	for ( int i = 0; i < Max_smcc_Layer ; i++ )
	{
		tcl.evalf ( "%s set dst_addr_ [Node allocaddr]", name() );
		smcc_addresses_smcc[i] = dst_.addr_;
		//tcl.evalf ( "%s join %d", name(), smcc_addresses_smcc[i] );
		printf ( " Sender %d joined address %d port %d\n",i, dst_.addr_ , dst_.port_ );

		clr_id[i] = -1;
		clr_timestamp[i] = 0;
		clr_last_feedback[i] = 0;
		round_id[i] = 0;
		if ( i == 0 )
			rate_[0] = InitRate_;
		else
			rate_[i] = 0;

	}

	Random::seed ( 10 );
	double now = Scheduler::instance().clock();
	for ( int i =0; i<Max_smcc_Layer; i++ )
	{
		send_timer_[i]= new SmccSendTimer ( this );
		NoFeedbacktimer_[i]= new SmccNoFeedbackTimer ( this );
	}
	for ( int i=0; i<Max_Bcl_Layer; i++ )
	{
		bcl_timer_[i]= new BclTimer ( this );
	}
//        send_timer_[Max_smcc_Layer]= SmccSendTimer(this);
//	printf(" SmccAgent::warmup()\n");
//        NoFeedbacktimer_[Max_smcc_Layer] = SmccNoFeedbackTimer(this);
//        bcl_timer_[Max_Bcl_Layer] =  BclTimer(this);


	for ( int i = 0; i < Max_smcc_Layer ; i++ )
	{

		seqno_[i] = 1;
		oldrate_[i] =  InitRate_;
		rate_change_ = INIT_RATE;
		last_change_[i] = 0;
		round_begin_[i] = now; // not 0, otherwise "no-feedback reduction" in first round
		maxrate_[i] = 0;
		expected_rate[i] = 0 ;
		inter_packet[i] = 0;
		join_next_layer[i] =0;
		max_rtt_[i] = Initial_RTT_;
		supp_rate_[i] = DBL_MAX;

		rtt_recv_id[i] = -1;
		rtt_prio[i] = 0;
		rtt_recv_timestamp[i] = 0;
		rtt_recv_last_feedback[i] = 0;
		rtt_rate[i] = 0;
		send_timer_[i]->smcc_layer = i;
		NoFeedbacktimer_[i]->smcc_layer = i;
		max_rate_on_layer[i] = base_layer_rate*pow ( 2,i ) ;
	}

	active_ = 1;
	bcl_round_seqno = 1;
	ndatapack_ = 0;
	delta_ = 0;
	highest_slow_start_smcc_layer =0;

	set_bcl_start();
//	printf(" SmccAgent::warmup()\n");

}

void SmccAgent::stop ( int layer_number )
{
	active_ = 0;
	send_timer_[layer_number]->force_cancel();
}

/*
 * Receive a status report from the receiver.
 */
void SmccAgent::recv ( Packet *pkt, Handler * )
{
//	printf("INSIDE RECV SENDER\n");
	double now = Scheduler::instance().clock();
	hdr_smcc_ack *nck = hdr_smcc_ack::access ( pkt );
	int cur_prio;
	int new_clr = 0;
	int smcc_layer = nck->smcc_layer;
	double adjusted_rate;
	if ( smcc_layer != 0 )
		join_next_layer[smcc_layer -1] = 0;


	if ( rate_change_ == INIT_RATE )
		rate_change_ = SLOW_START;


	// if we don't get ANY feedback for some time, cut rate in half
//	NoFeedbacktimer_[smcc_layer].resched(DEC_NO_REPORT * max_rtt_[smcc_layer]);

	// compute receiver's current RTT
	double tmp_rtt = now - nck->timestamp_echo - nck->timestamp_offset;
	if ( max_rtt_[smcc_layer] < tmp_rtt )
		max_rtt_[smcc_layer] = tmp_rtt;

	adjusted_rate = nck->rate;
	if ( nck->have_rtt == 0 && rate_change_ == CONG_AVOID )
	{
		// recalculate rate with correct RTT
		adjusted_rate = adjusted_rate / tmp_rtt * max_rtt_[smcc_layer];
	}

	if ( adjusted_rate < min_rate_ )
		adjusted_rate = min_rate_;

	// adjust suppression rate by G_FACTOR if non-CLR (and correct feedback round)
	// before clr_id is updated so that new CLR also contributes to suppression
	if ( ( clr_id[smcc_layer] != nck->recv_id ) && ( supp_rate_[smcc_layer] > ( 1-G_FACTOR ) * nck->rate ) &&
	        ( nck->round_id == round_id[smcc_layer] ) )
		supp_rate_[smcc_layer] = ( 1-G_FACTOR ) * nck->rate;

	// if we are in slow start and we just saw a loss then come out of slow start
	if ( rate_change_ == SLOW_START && nck->have_loss )
	{
		rate_change_ = CONG_AVOID;
//		printf(" at %f @@@@@@ adjusted rate %f\n", now,adjusted_rate *8/1000000);
		clr_id[smcc_layer] = nck->recv_id;
		new_clr = 1;
		double temp = adjusted_rate;
		for ( int i=0; i < Max_smcc_Layer ; i++ )
		{
			if ( temp > max_rate_on_layer[i] )
			{
				rate_[i] = max_rate_on_layer[i] ;
				if ( oldrate_[i] == InitRate_ )
					create_new_smcc_layer ( i-1 ) ;
			}
			else
			{
				rate_[i] = temp ;
				break;
			}
		}
	}
	else
	{

		if ( rate_change_ == SLOW_START )
		{
			// adjust maxrate
			if ( clr_id[smcc_layer] == nck->recv_id )
			{
				maxrate_[smcc_layer] = adjusted_rate;
			}
			else if ( ( nck->receiver_leave == 0 ) && ( adjusted_rate < maxrate_[smcc_layer] || clr_id[smcc_layer] == -1 ) )
			{
				// change CLR
				maxrate_[smcc_layer] = adjusted_rate;
				clr_id[smcc_layer] = nck->recv_id;
				new_clr = 1;
			}
			if ( smcc_layer == highest_slow_start_smcc_layer+1 )
				highest_slow_start_smcc_layer++;
			slowstart ( highest_slow_start_smcc_layer );

		}
		else if ( rate_change_== CONG_AVOID )
		{
			if ( ( nck->receiver_leave == 0 ) && ( nck->have_loss ) &&
			        ( adjusted_rate < rate_[smcc_layer] || clr_id[smcc_layer] == -1 ) )
			{
				// change CLR
				if ( ( clr_id[smcc_layer] == -1 ) && ( rate_[smcc_layer] != 0 ) )
				{
					if ( NoFeedbacktimer_[smcc_layer]->status() == TIMER_PENDING )
					{
						printf ( "at %f @@@ S received when clr is -1 from smcc layer %d\n",
						         now,smcc_layer );
						NoFeedbacktimer_[smcc_layer]->cancel();
						if ( ( smcc_layer ) && ( rate_[smcc_layer] == max_rate_on_layer[smcc_layer -1] ) &&
						        ( adjusted_rate == max_rate_on_layer[smcc_layer] ) )
						{
							rate_[smcc_layer] = max_rate_on_layer[smcc_layer] ;
							printf ( "at %f @@@ S set rate on L %d to %f\n",now,
							         smcc_layer,rate_[smcc_layer]*8/1000000 );
						}
					}

				}
				new_clr = 1;

				if ( nck->recv_id != clr_id[smcc_layer] )
				{
					printf ( " @@ rate %f received rate %f previoud clr %d\n",
					         rate_[smcc_layer]*8/1000000,adjusted_rate*8/1000000,clr_id[smcc_layer] );
					printf ( " @@ CLR Change at %f on L %d to  %d \n",now,smcc_layer,nck->recv_id );
				}
				clr_id[smcc_layer] = nck->recv_id;
				// clr on lower layer becomes clr on this layer from join attempt.
				// make clr on lower layer to -2
				if ( clr_id[smcc_layer -1] == nck->recv_id )
					clr_id[smcc_layer - 1] = -2;
			}

			// when the sender receives the leave message from the clr.
			if ( ( nck->receiver_leave == 1 ) && ( clr_id[smcc_layer] == nck->recv_id ) )
			{
				clr_id[smcc_layer] = -1;
				printf ( "at %f @@@ S received leave msg from smcc layer %d\n",now,smcc_layer );
				expected_rate[smcc_layer] = adjusted_rate;
				oldrate_[smcc_layer] = rate_[smcc_layer ] = max_rate_on_layer[smcc_layer -1];
				last_change_[smcc_layer] = now;
				round_id[smcc_layer] = ( round_id[smcc_layer] + 1 ) % MAX_ROUND_ID;
				round_begin_[smcc_layer] = now;
				supp_rate_[smcc_layer] = DBL_MAX;
				sendpkt ( smcc_layer );
				NoFeedbacktimer_[smcc_layer]->resched ( ( DEC_NO_REPORT+4 ) *max_rtt_[smcc_layer] );
			}

			if ( clr_id[smcc_layer] == nck->recv_id )
			{
				if ( adjusted_rate >= 0 )
				{
					expected_rate[smcc_layer] = adjusted_rate;
					if ( ( smcc_layer ) && ( expected_rate[smcc_layer] < max_rate_on_layer[smcc_layer -1] ) )
					{
						printf ( " ERROR \n" );
						exit ( 0 );
					}

					if ( expected_rate[smcc_layer] >= rate_[smcc_layer] )
						increase_rate ( smcc_layer );
					else
						decrease_rate ( smcc_layer );

				}
			}
		}
	}

	// determine which timestamp to include in data packet
	if ( clr_id[smcc_layer] == nck->recv_id )
	{
		// remember CLR timestamp to include in packets w/o any other RTT feedback
		clr_timestamp[smcc_layer] = nck->timestamp;
		clr_last_feedback[smcc_layer] = now;
		if ( !nck->have_rtt || new_clr )
			cur_prio = PRIO_FORCE;
		else
			cur_prio = PRIO_CLR;
	}
	else
	{
		if ( !nck->have_rtt )
			cur_prio = PRIO_NO_RTT;
		else
			cur_prio = PRIO_RECV;
	}
	if ( cur_prio > rtt_prio[smcc_layer] || ( cur_prio == rtt_prio[smcc_layer]
	        && nck->rate < rtt_rate[smcc_layer] ) )
	{
		rtt_prio[smcc_layer] = cur_prio;
		rtt_recv_id[smcc_layer] = nck->recv_id;
		rtt_recv_timestamp[smcc_layer] = nck->timestamp;
		rtt_recv_last_feedback[smcc_layer] = now;
		rtt_rate[smcc_layer] = nck->rate; // use reported rate for suppression
	}


	if ( printStatus_ )
	{
		int i = smcc_layer;
		if ( nck->receiver_leave )
			printf ( "s%i on L %d %f leave-group", sndr_id,i, now );
		else
			printf ( "s%i on L %d %f update", sndr_id,i, now );
		if ( supp_rate_[i] == DBL_MAX )
		{
			printf ( " [%i,i%i,m%i] %f/%f sendrate %f (supp oo, max %f, max rtt %f) round %i(%i) ",
			         nck->recv_id, rtt_recv_id[i], clr_id[i], adjusted_rate*8.0  / 1000000.0,
			         nck->rate*8.0  / 1000000.0,
			         rate_[i]*8.0  / 1000000.0, maxrate_[i]*8.0  / 1000000.0,
			         max_rtt_[i], round_id[i], nck->round_id );
		}
		else
		{
			printf ( " [%i,i%i,m%i] %f/%f sendrate %f (supp %f, max %f, max rtt %f) round %i(%i) ",
			         nck->recv_id, rtt_recv_id[i], clr_id[i], adjusted_rate*8.0  / 1000000.0,
			         nck->rate*8.0  / 1000000.0,
			         rate_[i]*8.0  / 1000000.0, supp_rate_[i]*8.0  / 1000000.0,
			         maxrate_[i]*8.0  / 1000000.0,
			         max_rtt_[i], round_id[i], nck->round_id );
		}
		printf ( ( nck->recv_id == clr_id[i] ) ? "!\n" : "\n" ); // feedback was from rep
	}
	Packet::free ( pkt );
}

void SmccAgent::slowstart ( int layer )
{
	double now = Scheduler::instance().clock();

	if ( rate_[layer]+SMALLFLOAT< size_/max_rtt_[layer] )
	{
		/* if this is the first report, change rate to 1 per rtt */
		/* compute delta so rate increases slowly to new value   */
		oldrate_[layer] = rate_[layer];
		rate_[layer] = fairsize_/max_rtt_[layer];
		delta_ = ( rate_[layer] - oldrate_[layer] ) / ( rate_[layer]*max_rtt_[layer]/size_ );
		last_change_[layer] = now;
		printf ( "22222\n" );
	}
	else
	{
		/* else multiply the rate by ssmult_, and compute delta, */
		/*  so that the rate increases slowly to new value       */
		if ( maxrate_ [layer]> 0 )
		{
			if ( ssmult_*rate_[layer] < maxrate_[layer] && now - last_change_[layer] > max_rtt_[layer] )
			{
				rate_[layer] = ssmult_*rate_[layer];
				delta_ = ( rate_[layer] - oldrate_[layer] ) / ( rate_[layer]*max_rtt_[layer]/size_ );
				last_change_[layer] = now;
			}
			else
			{
				if ( rate_[layer] > maxrate_[layer] || now - last_change_[layer] > max_rtt_[layer] )
				{
					rate_[layer] = maxrate_[layer];
					delta_ = ( maxrate_[layer] - oldrate_[layer] ) / ( rate_[layer]*max_rtt_[layer]/size_ );
					if ( delta_ < 0 ) delta_ = 0;
					last_change_[layer] = now;
				}
			}
		}
		else
		{
			if ( now - last_change_[layer] > max_rtt_[layer] )
			{
				rate_[layer] = ssmult_*rate_[layer];
				delta_ = ( rate_[layer] - oldrate_[layer] ) / ( rate_[layer]*max_rtt_[layer]/size_ );
				last_change_[layer] = now;
			}
		}
		if ( ( oldrate_[layer]+SMALLFLOAT< rate_[layer] ) )
		{
			oldrate_[layer] = oldrate_[layer] + delta_;
			if ( layer )
			{
				double diff_rate =  oldrate_[layer] - max_rate_on_layer[layer-1];
				double next = size_/diff_rate;
				if ( inter_packet[layer] >= 2 * next )
					if ( next_send_time > ( next + now ) )
					{
						send_timer_[layer]->resched ( next );
						printf ( "@@ SLOW START check if rate increase L %d new resched  at %f next %f,previ %f\n",
						         layer,next+now,next,inter_packet[layer] );
						next_send_time =next + now;
					}
			}
		}
	}
}

void SmccAgent::increase_rate ( int layer_number )
{
	double now = Scheduler::instance().clock();
	double mult = ( now - last_change_[layer_number] ) / max_rtt_[layer_number];
	if ( mult > 2 ) mult = 2;

	if ( rate_[layer_number] == 0 )
	{
		rate_[layer_number] = max_rate_on_layer[layer_number -1];
		printf ( " at %f @@@ new layer with new report rate %f on L %d\n",
		         now, rate_[layer_number]*8/1000000,layer_number );
		double diff_rate =  rate_[layer_number] - rate_[layer_number -1];
		printf("rate_[layer_number] %f - rate_[layer_number -1 %f = %f\n",rate_[layer_number],rate_[layer_number -1],diff_rate);
		if(diff_rate == 0.0) diff_rate = 0.1*rate_[layer_number];
		inter_packet[layer_number] = size_/diff_rate;
		last_change_[layer_number] = now;
		max_rtt_[layer_number] = max_rtt_[layer_number -1] ;
		// randomize between +-50% packet interval, but no more than max_rtt gap
		double max_variation = MAX ( 0, MIN ( 0.5 * inter_packet[layer_number],
		                                      max_rtt_[layer_number] - inter_packet[layer_number] ) );
		double next = Random::uniform ( inter_packet[layer_number] - max_variation,
		                                inter_packet[layer_number] + max_variation );
		assert ( next > SMALLFLOAT );
		join_next_layer[layer_number] = 0;
		if ( next  < 0 )
			printf ( " ~~~~~~~ ERROR  ~~~~ 1\n" );
		send_timer_[layer_number]->resched ( next );
		next_send_time = now + next;
	}
	else
		rate_[layer_number] = MIN ( expected_rate[layer_number],
		                            rate_[layer_number] + ( fairsize_/max_rtt_[layer_number] ) *mult );


	if ( rate_[layer_number] < fairsize_/max_rtt_[layer_number] )
		rate_[layer_number] = fairsize_/max_rtt_[layer_number];

	last_change_[layer_number] = now;

	if ( rate_[layer_number] > max_rate_on_layer[layer_number] )
	{

		if ( ( rate_[layer_number+1] == 0 ) || ( clr_id[layer_number + 1] == -1 ) )   // there is no receiver on next layer
		{
			create_new_smcc_layer ( layer_number ) ;
			printf ( "@@ Create new layer \n" );
			clr_id[layer_number+1] = clr_id[layer_number] ;
			rtt_recv_id[layer_number+1] =  clr_id[layer_number] ; //  send new clr info in new layer
			clr_id[layer_number] = -2 ;  //  no active receiver on this layer
			if ( NoFeedbacktimer_[layer_number+1]->status() == TIMER_PENDING )
			{
				printf ( "at %f @@@ S Candel Nofeedback timer on smcc layer %d\n",now,layer_number+1 );
				NoFeedbacktimer_[layer_number+1]->cancel();
			}
		}
		else
			send_pkt_at_max_rate_smcc_layer ( layer_number );

	}
	if ( ( layer_number ) && ( rate_[layer_number] < max_rate_on_layer[layer_number -1] ) )
	{
		printf ( " rate %f  max rate of low layer %f on L %d \n",  rate_[layer_number]*8/1000000,max_rate_on_layer[layer_number -1]*8/1000000, layer_number );
		exit ( 0 );
	}

	// check if rate increase should impact send timer
	if ( layer_number )
	{
		double diff_rate =  rate_[layer_number] - rate_[layer_number -1];
		if(diff_rate == 0.0) diff_rate = 0.1*rate_[layer_number];
		double next = size_/diff_rate;
		if ( next < 0 ) printf ( " ~~~~~~~ ERROR \n" );
		if ( inter_packet[layer_number] >= 2 * next )
		{
			assert ( next > SMALLFLOAT );

			if ( next  < 0 )
				printf ( " ~~~~~~~ ERROR ~~~~ 4\n" );
			if ( next_send_time > ( next + now ) )
			{
				send_timer_[layer_number]->resched ( next );
				printf ( "check if rate increase L %d new resched  at %f next %f\n",layer_number,next+now,next );
				next_send_time =next + now;
			}
		}
	}
}

void SmccAgent::decrease_rate ( int layer_number )
{
	double now = Scheduler::instance().clock();
	oldrate_[layer_number ] = rate_[layer_number ] = expected_rate[layer_number];
	last_change_[layer_number] = now;
}
void SmccAgent::create_new_smcc_layer ( int layer_number )
{
	
	double now = Scheduler::instance().clock();

	rate_[layer_number+1] = rate_[layer_number] ;
	rate_[layer_number] = max_rate_on_layer[layer_number];
	double diff_rate =  rate_[layer_number+1] - rate_[layer_number];
	if(diff_rate ==0) return;
	printf("rate_[layer_number] %f - rate_[layer_number -1] %f = %f\n",rate_[layer_number+1],rate_[layer_number],diff_rate);
	printf("SmccAgent::create_new_smcc_layer diff_rate %f \n", diff_rate);
	clr_id[layer_number+1] = clr_id[layer_number] ;
	oldrate_[layer_number+1] = oldrate_[layer_number] ;
	if ( rate_change_ == SLOW_START )
		diff_rate =  oldrate_[layer_number+1] - max_rate_on_layer[layer_number];
	
	
	inter_packet[layer_number+1] = size_/diff_rate;
	last_change_[layer_number+1] = last_change_[layer_number] ;
	max_rtt_[layer_number+1] = max_rtt_[layer_number] ;

	// randomize between +-50% packet interval, but no more than max_rtt gap
	double max_variation = MAX ( 0, MIN ( 0.5 * inter_packet[layer_number+1],
	                                      max_rtt_[layer_number+1] - inter_packet[layer_number+1] ) );
	double next = Random::uniform ( inter_packet[layer_number+1] - max_variation,
	                                inter_packet[layer_number+1] + max_variation );
	assert ( next > SMALLFLOAT );
	join_next_layer[layer_number] = 1;
	if ( diff_rate <0 ) printf ( " ~~~~~~~ ERROR  2\n" );
	printf ( "rate %f new rate %f next %f \n",
	         rate_[layer_number]*8/1000000, rate_[layer_number+1]*8/1000000,next );

	if ( next  < 0 )
		printf ( " ~~~~~~~ ERROR ~~~~ 2\n" );


	send_timer_[layer_number+1]->resched ( next );
	next_send_time =next + now;
	printf ( "at %f @@@ ** NEW LAYER  **\n",now );

}

void SmccAgent::send_pkt_at_max_rate_smcc_layer ( int layer_number )
{
//	double now = Scheduler::instance().clock();
	rate_[layer_number] = max_rate_on_layer[layer_number];
}

void SmccAgent::nextpkt ( int layer_number )
{
	double now = Scheduler::instance().clock();
	double xrate = -1;
	double next2;
	// new feedback round every T = t_factor * max_RTT,
	//   if at least one non-CLR FBM was received (supp_rate_ < DBL_MAX)
	//   or after T = 2 * t_factor * max_RTT (only one receiver in group)
	if ( ( now - round_begin_[layer_number] >= t_factor_ * max_rtt_[layer_number] ) && ( supp_rate_[layer_number] < DBL_MAX ) ||
	        ( now - round_begin_[layer_number] >= 2 * t_factor_ * max_rtt_[layer_number] ) )
	{
		round_id[layer_number] = ( round_id[layer_number] + 1 ) % MAX_ROUND_ID;
		round_begin_[layer_number] = now;
		supp_rate_[layer_number] = DBL_MAX;
		// XXX draft says only inc when max_rtt unchanged from prev round
		max_rtt_[layer_number] *= MAX_RTT_DECAY;
		if ( max_rtt_[layer_number] < size_/rate_[layer_number] )
			max_rtt_[layer_number] = size_/rate_[layer_number];
	}

	// time out CLR when there is an active receiver on layer
	// clr_id[i] is -2 when there is no active receiver on layer i
	if ( ( clr_id[layer_number] != -2 ) &&
	        ( now - clr_last_feedback[layer_number] > max_rtt_[layer_number] * DEC_NO_REPORT ) )
	{
		clr_id[layer_number] = -1;
	}

	sendpkt ( layer_number );

	// during slow start, we increase rate
	// slowly - by amount delta per packet
	if ( ( rate_change_ == SLOW_START ) && ( oldrate_[layer_number]+SMALLFLOAT< rate_[layer_number] ) )
	{
		xrate = oldrate_[layer_number];
	}
	else
	{
		xrate = rate_[layer_number];
	}


	if ( xrate > 0 )
	{
		if ( ( xrate >= max_rate_on_layer[layer_number] ) && ( rate_change_ == SLOW_START ) && ( oldrate_[layer_number+1] == InitRate_ ) )
		{
			maxrate_[layer_number+1]= maxrate_[layer_number];
			last_change_[layer_number+1] = last_change_[layer_number] ;
			oldrate_[layer_number+1] = oldrate_[layer_number];
			max_rtt_[layer_number+1] = max_rtt_[layer_number];
//			printf(" @@@ Call create new layer:  rate %f , max_rate %f\n",
//				xrate*8/1000000.0,max_rate_on_layer[layer_number]*8/1000000);
			create_new_smcc_layer ( layer_number ) ;
		}

		double diff_rate = xrate ;
		if ( ( rate_change_ == SLOW_START ) && ( layer_number ) )
		{
			diff_rate = diff_rate - oldrate_[layer_number -1];
		}
		else if ( layer_number )
		{
			for ( int i = layer_number -1 ; i >=0 ; i-- )
			{
				if ( rate_[i] > 0 )
				{
					diff_rate = diff_rate - rate_[i] ;
					break;
				}
			}
		}
		if ( ( layer_number ) && ( rate_change_ != SLOW_START ) )
			if ( xrate < rate_[layer_number -1 ] )
			{
				printf ( " rate is smaller than lower xrate %f, rate %f\n",
				         xrate*8/1000000,rate_[layer_number-1]*8/1000000 );
				exit ( 0 );
			}

		if ( diff_rate > SMALLFLOAT )
		{
			inter_packet[layer_number] = size_/diff_rate;

			// randomize between +-50% packet interval, but no more than max_rtt gap
			double max_variation = MAX ( 0, MIN ( 0.5 * inter_packet[layer_number],
			                                      max_rtt_[layer_number] - inter_packet[layer_number] ) );
			double next = Random::uniform ( inter_packet[layer_number] - max_variation,
			                                inter_packet[layer_number] + max_variation );
			assert ( next > SMALLFLOAT );
			if ( next <  0 )
			{
				printf ( " ~~~~~~~ ERROR ~~~~ 5\n" );
				exit ( 0 );
			}
			send_timer_[layer_number]->resched ( next );
			next2 = next+now;
			next_send_time = next + now;
		}

	}
}

void SmccAgent::sendpkt ( int layer_number )
{
	if ( active_ )
	{
		double now = Scheduler::instance().clock();
		dst_.addr_ = smcc_addresses_smcc[layer_number];
		Packet* p = allocpkt();
		hdr_smcc *smcch = hdr_smcc::access ( p );
		smcch->timestamp = now;
		smcch->rate_change =rate_change_;
		smcch->supp_rate = supp_rate_[layer_number];
		smcch->join_next_layer = join_next_layer[layer_number];
		smcch->next_layer_rate = rate_[layer_number+1];
		smcch->sending_rate = rate_[layer_number];
		smcch->psize = size_;
		smcch->packet_type = SMCC_Type;
		smcch->round_id = round_id[layer_number];
		smcch->round_num_bcl= round_num_bcl;
		smcch->max_rtt = max_rtt_[layer_number];
		smcch->smcc_layer = layer_number;
		smcch->seqno=seqno_[layer_number]++;


//printf("######SMCC Sendpkt time %f layer_number %d seqno %d \n", smcch->timestamp,layer_number,smcch->seqno);


		for ( int i = 0; i < Max_smcc_Layer ; i++ )
		{
			smcch->clr_id[i] =clr_id[i];
		}


		if ( clr_id[layer_number] == rtt_recv_id[layer_number] || rtt_recv_id[layer_number] == -1
		        || clr_id[layer_number] == -1 )
		{
			smcch->rtt_recv_id = clr_id[layer_number];
			smcch->timestamp_echo =  clr_timestamp[layer_number];
			smcch->timestamp_offset = now - clr_last_feedback[layer_number];
			smcch->is_clr = 1;
		}
		else
		{
			smcch->rtt_recv_id = rtt_recv_id[layer_number];
			smcch->timestamp_echo = rtt_recv_timestamp[layer_number];
			smcch->timestamp_offset = now - rtt_recv_last_feedback[layer_number];
			smcch->is_clr = 0;
		}

		//printf("before send(p,0) \n");
		send ( p, 0 );
		// target_->recv(p,(Handler *)0);


		if ( 0 )
		{
			int i = layer_number;
			if ( supp_rate_[i] == DBL_MAX )
			{
				printf ( "s%i %f data-packet [i%i,m%i] sendrate %f (supp oo, max %f, max rtt %f) round %i\n",
				         sndr_id, now, rtt_recv_id[i], clr_id[i],
				         rate_[i] * 8.0 / 1000.0, maxrate_[i] * 8.0 / 1000.0,
				         max_rtt_[i], round_id[i] );
			}
			else
			{
				printf ( "s%i %f data-packet [i%i,m%i] sendrate %f (supp %f, max %f, max rtt %f) round %i\n",
				         sndr_id, now, rtt_recv_id[i], clr_id[i],
				         rate_[i] * 8.0 / 1000.0, supp_rate_[i] * 8.0 / 1000.0,
				         maxrate_[i] * 8.0 / 1000.0, max_rtt_[i], round_id[i] );
			}
		}
		ndatapack_++;
		rtt_recv_id[layer_number] = -1;
		rtt_prio[layer_number] = 0;

	}
}
void SmccAgent::reduce_rate_to_zero_on_no_feedback ( int layer_number )
{
	rate_[layer_number] = 0;
	printf ( "@@@ s%i %f no-feedback reduction set rate to 0 on layer %d\n",  sndr_id, Scheduler::instance().clock(),layer_number );

}

void SmccAgent::reduce_rate_on_no_feedback ( int layer_number )
{
	rate_[layer_number] *= 0.5;
	double next =  size_ / rate_[layer_number];
	assert ( next > SMALLFLOAT );
	send_timer_[layer_number]->resched ( next );

	NoFeedbacktimer_[layer_number]->resched ( DEC_NO_REPORT*max_rtt_[layer_number] );

	if ( printStatus_ )
		printf ( "s%i %f no-feedback reduction\n", sndr_id, Scheduler::instance().clock() );
}

void SmccSendTimer::expire ( Event * )
{
	a_->nextpkt ( smcc_layer );

}

void SmccNoFeedbackTimer::expire ( Event * )
{
	a_->reduce_rate_to_zero_on_no_feedback ( smcc_layer );
//	a_->reduce_rate_on_no_feedback(smcc_layer);
}

void BclTimer::expire ( Event * )
{
	a_->bcl_timeout ( layer_number_,join_next_bcl );
}

void SmccAgent::bcl_timeout ( int layer_number,int join_next_bcl )
{

	get_bcl_interval ( layer_number );

	bcl_sendit ( layer_number,join_next_bcl );
}


void SmccAgent::bcl_sendit ( int layer_number,int join_next_bcl )
{

//        Tcl & tcl = Tcl::instance();
	dst_.addr_ = bcl_addresses_smcc[layer_number];

//	printf("bcl_sendit address %d layer_number %d \n", dst_.addr_, layer_number);
	Packet* p = allocpkt();
//        hdr_cmn* hdrcmn = hdr_cmn::access(p);
	hdr_ip* hdrip = hdr_ip::access ( p );
	double now = Scheduler::instance().clock();

	hdr_smcc *smcch = hdr_smcc::access ( p );
	smcch->seqno = seq_number_bcl[layer_number];
	smcch->timestamp = now;
	smcch->packet_type = BCL_Type; // BCL
	smcch->psize=size_;
	smcch->round_num_bcl= rtt_round_seqno;
	smcch->join_next_bcl= join_next_bcl;
	round_num_bcl= rtt_round_seqno;
	smcch->bcl_layer = layer_number;
	hdrip->fid_ = layer_number + 80;
// 	hdr_cmn* hdr = hdr_cmn::access ( p );
//	printf("SMCC time %f layer_number %d seqno %d \n", smcch->timestamp,layer_number,smcch->seqno);
// 	printf("SENDER CMN time %f packet_id %d \n", hdr->timestamp(),hdr->uid_);
//	printf("SmccAgent::bcl_sendit sdr_port %d dst_.port_ %d\n", hdrip->sport(),  hdrip->dport());
	target_->recv ( p, ( Handler * ) 0 );



}


void SmccAgent::set_bcl_start( )
{
//        double now = Scheduler::instance().clock();
	bcl_addresses_smcc = new nsaddr_t[Max_Bcl_Layer];
	rtt_round_seqno = 0;
	Tcl & tcl = Tcl::instance();
	for ( int i = 0; i < Max_Bcl_Layer ; i++ )
	{
		bcl_timer_[i]->counter_ = 1;
		bcl_timer_[i]->start_phase_ = 1;
		bcl_timer_[i]->join_next_bcl = 0;
		if ( i !=0 )
			bcl_timer_[i]->interval_ = min_rtt_smcc_ + min_rtt_smcc_/ ( pow ( 2,i+2 ) );
		else
			bcl_timer_[i]->interval_ = min_rtt_smcc_;

		seq_number_bcl[i] = 0;
		bcl_timer_[i]->layer_number_ = i;

		tcl.evalf ( "%s set dst_addr_ [Node allocaddr]", name() );
		bcl_addresses_smcc[i] = dst_.addr_  ;
		//tcl.evalf ( "%s join %d", name(), bcl_addresses_smcc[i] );
		printf ( " %d bcl_addresses_smcc %d port %d\n", i, dst_.addr_ ,dst_.port_ );
		//	printf("SmccAgent::set_bcl_start( )\n");
		bcl_timer_[i]->resched ( bcl_timer_[i]->interval_ );
	}
}


void SmccAgent::get_bcl_interval ( int layer_number_ )
{
	// counter_ is the sequential number of packet in the cycle
//        double now = Scheduler::instance().clock();
	switch ( layer_number_ )
	{
		case 0 :
			switch ( bcl_timer_[layer_number_]->counter_ )
			{
				case 1 :
					++rtt_round_seqno;
					bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_;
					bcl_timer_[layer_number_]->counter_++;
					break;
				case 2 :
					++rtt_round_seqno;
					bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 2;
					bcl_timer_[layer_number_]->counter_++;
					break;
				case 3 :
					bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 2;
					bcl_timer_[layer_number_]->counter_ = 1;
					break;
			} // end of switch
			if ( bcl_timer_[layer_number_]->counter_== 3 )
				bcl_timer_[layer_number_]->join_next_bcl = 1;
			else
				bcl_timer_[layer_number_]->join_next_bcl = 0;
			break; // end of case 0
		case 1 :
			switch ( bcl_timer_[layer_number_]->counter_ )
			{
case 1 : case 2: case 3: case 4: case 5: case 6: case 7:
				case 8:
					bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 4;
					bcl_timer_[layer_number_]->counter_++;
					break;
		case 9 : case 10: case 11:
					bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 2;
					bcl_timer_[layer_number_]->counter_++;
					break;
				case 12:
					bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 2;
					bcl_timer_[layer_number_]->counter_ = 1;
					break;
			} // end of switch
			if ( bcl_timer_[layer_number_]->counter_== 8 )
				bcl_timer_[layer_number_]->join_next_bcl = 1;
			else
				bcl_timer_[layer_number_]->join_next_bcl = 0;
			break;
		case 2 :
			if ( bcl_timer_[layer_number_]->start_phase_ )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 4;
				if ( bcl_timer_[layer_number_]->counter_ ==  8 )
				{
					bcl_timer_[layer_number_]->counter_ = 1;
					bcl_timer_[layer_number_]->start_phase_ = 0;
				}
				else
					bcl_timer_[layer_number_]->counter_++;
				break;
			}
			if ( bcl_timer_[layer_number_]->counter_ <= 32 )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 8;
				bcl_timer_[layer_number_]->counter_++;
			}
			else if ( bcl_timer_[layer_number_]->counter_ < 48 )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 4;
				bcl_timer_[layer_number_]->counter_++;
			}
			else if ( bcl_timer_[layer_number_]->counter_ ==  48 )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / 4;
				bcl_timer_[layer_number_]->counter_ = 1;
			}
			if ( bcl_timer_[layer_number_]->counter_== 32 )
				bcl_timer_[layer_number_]->join_next_bcl = 1;
			else
				bcl_timer_[layer_number_]->join_next_bcl = 0;
			break;
		default :
			if ( bcl_timer_[layer_number_]->start_phase_ )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / pow ( 2,layer_number_ );
				int temp, temp2;
				temp = ( int ) pow ( 2,layer_number_ ) - 2 ;
				temp2 = temp * ( int ) pow ( 2,layer_number_ );
				if ( bcl_timer_[layer_number_]->counter_ ==  temp2 )
				{
					bcl_timer_[layer_number_]->counter_ = 1;
					bcl_timer_[layer_number_]->start_phase_ = 0;
				}
				else
					bcl_timer_[layer_number_]->counter_++;
				break;
			}
			int temp3,temp4,temp5;
			temp3 = ( int ) pow ( 2,layer_number_ );
			temp4 = temp3 * ( int ) pow ( 2,layer_number_ + 1 );
			temp5 = temp4 + ( temp3*temp3 );
			if ( bcl_timer_[layer_number_]->counter_ <= temp4 )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / pow ( 2,layer_number_ + 1 );
				bcl_timer_[layer_number_]->counter_++;
			}
			else if ( bcl_timer_[layer_number_]->counter_ < temp5 )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / temp3;
				bcl_timer_[layer_number_]->counter_++;
			}
			else if ( bcl_timer_[layer_number_]->counter_ ==  temp5 )
			{
				bcl_timer_[layer_number_]->interval_ = min_rtt_smcc_ / temp3;
				bcl_timer_[layer_number_]->counter_ = 1;
			}
			if ( bcl_timer_[layer_number_]->counter_== temp4 )
				bcl_timer_[layer_number_]->join_next_bcl = 1;
			else
				bcl_timer_[layer_number_]->join_next_bcl = 0;
			break;
	} // end of switch
	seq_number_bcl[layer_number_]++;
	if ( bcl_timer_[layer_number_]->interval_  < 0 )
		printf ( " ~~~~~~~ ERROR ~~~~ 7\n" );
	bcl_timer_[layer_number_]->resched ( bcl_timer_[layer_number_]->interval_ );
}







static class SmccSinkClass : public TclClass
{
	public:
		SmccSinkClass() : TclClass ( "Agent/SMCCSink" ) {}
		TclObject* create ( int, const char*const* )
		{
			return ( new SmccSinkAgent() );
		}
} class_smccSink;


SmccSinkAgent::SmccSinkAgent() : Agent ( PT_SMCC_ACK )
{
	bind ( "packetSize_", &size_ );
	bind ( "fairSize_", &fairsize_ );
	bind_bool ( "aggregatePackets_", &aggregate_packets_ );
	bind_bool ( "aggregateLoss_", &aggregate_loss_ );
	bind ( "InitHistorySize_", &hsz );
	bind ( "AdjustHistoryAfterSS_", &adjust_history_after_ss );
	bind ( "NumSamples_", &numsamples );
	bind ( "discount_", &discount );
	bind ( "printStatus_", &printStatus_ );
	bind ( "smooth_", &smooth_ );
	bind ( "recv_id_", &recv_id );
	bind ( "df_", &df_ );
	bind ( "rate_", &rate_ );
	bind ( "bval_", &bval_ );
	bind ( "tight_loop_", &tight_loop_ );
	bind ( "fbtime_mult_", &fbtime_mult_ );
	bind ( "constant_rate_", &constant_rate_ );
	bind ( "display_rate_", &display_rate_ );
	bind ( "total_bytes_", &total_bytes_ );
	bind ( "base_layer_rate", &base_layer_rate );
	bind ( "damping_factor", &damping_factor );
	bind ( "without_join_attempt", &without_join_attempt );
	//GK
	bind("end_2_end_delay_",&end_2_end_delay_);
	bind("jitter_",&jitter_);
	

	// XXX some of this should be set in start proc (which doesn't exist yet)

	rtt_ = 0;
	rtt_est = 0;
	one_way_ = 0;
	last_timestamp_ = 0;
	last_arrival_ = 0;
	last_arrival_layer = 0;
	display_rate_=0;
	total_bytes_=0;
	avgsize_ = 0;
	sum_psize_ = 0;
	sum_lsize_ = 0;
	bcl_join_pending = 0;
	rate_change = SLOW_START;
	got_request_for_new_clr_on_this_layer = -1;

	active_ = 1;
	rcvd_since_last_report  = 0;
	loss_seen_yet = 0;
	lastloss = -1000; // also count loss event directly after startup
	false_sample = 0;
	expected_num_packet_from_bcl = 0;
	sample_count = 1 ;
	mult_factor_ = 1.0;
	representative = 0;
	comparison_rate = 0;
	suppression_rate = 0;
	flag_join_bcl = 0;
	receiver_leave = 0;
	Random::seed ( 20 );


	sample = NULL ;
	weights = NULL ;
	mult = NULL ;
	highest_smcc_layer = 0;
	highest_bcl_layer =  -1;
	join_pending = 0;
	leave_pending = 0;
	next_layer_rate = 0;
	last_report_sent =0;
	for ( int i= 0; i< Max_smcc_Layer ; i++ )
	{
		receiving_rate[i] =  0;
		round_id[i] = -1;    // so that first data packet starts a new round
		fb_time = 0;
		rtvec_[i] = NULL;
		tsvec_[i] = NULL;
		lossvec_[i] = NULL;
		sizevec_[i] = NULL;
		maxseq[i] = -1;
		last_sample[i] = 0;
	}
	ptransit_ =0;
	// for statistics
	total_pkts_recv_ = total_pkts_lost_ = 0;
	
	
	


}



/*
 * Receive new data packet.  If appropriate, generate a new report.
 */
void SmccSinkAgent::recv ( Packet *pkt, Handler * )
{
//	printf ( "SmccSinkAgent::recv\n" );
	hdr_smcc *smcch = hdr_smcc::access ( pkt );
	double now = Scheduler::instance().clock();
	double p = -1;
	double rtt_cur;
	int new_round;
	double prev_last_arrival_;
	display_rate_+=  hdr_cmn::access ( pkt )->size();
	total_bytes_+=  hdr_cmn::access ( pkt )->size();
	round_num_bcl = smcch->round_num_bcl;
	int join_next_layer = smcch->join_next_layer;
	layer_number = smcch->smcc_layer;
	int prevmaxseq = maxseq[layer_number];
	last_arrival_ = now;
	
	// GK
	total_pkts_recv_++;
	end_2_end_delay_ = now - smcch->timestamp;
	// measure  jitter
	// calculate delay jitter, see RFC 3550 Appendix A8
 		double transit = now - smcch->timestamp;
 		double d = transit - ptransit_;
 		if ( d < 0 ) d = -d;
		double temp = d- jitter_;
		if(temp <0) temp = -temp;
 		jitter_ = (jitter_ + (1./16.) ) * (temp);
// 		pjitter = jitter_;
		//s->jitter ( s->jitter() + ( 1./16. ) * ( d - s->jitter() ) );	
	
//	printf("######SMCC RECEIVE time %f layer_number %d seqno %d  ##########\n", smcch->timestamp,layer_number,smcch->seqno);


	if ( layer_number  > highest_smcc_layer )
		return;

	receiving_rate[layer_number] = smcch->sending_rate;
	if ( ( layer_number == highest_smcc_layer ) && ( smcch->packet_type == SMCC_Type ) )
	{
		next_layer_rate = smcch->next_layer_rate;
		if ( join_pending )
		{
			round_id[layer_number] = smcch->round_id;
			new_round = 1;
		}
	}

	double Mrate ;

	if ( smcch->packet_type == BCL_Type )
	{
		bcl_recv ( pkt );
		return;
	}

	rcvd_since_last_report ++;

	if ( numsamples < 0 )
	{
		// This is the first packet received.
		numsamples = DEFAULT_NUMSAMPLES ;
		// forget about losses before this
		prevmaxseq = maxseq[0] = smcch->seqno-1 ;
		// initialize last_sample
		last_sample[0] = smcch->seqno;

		if ( smooth_ == 1 )
		{
			numsamples = numsamples + 1;
		}
		sample = ( int * ) malloc ( ( numsamples+1 ) *sizeof ( int ) );
		weights = ( double * ) malloc ( ( numsamples+1 ) *sizeof ( double ) );
		mult = ( double * ) malloc ( ( numsamples+1 ) *sizeof ( double ) );
		for ( int i = 0 ; i < numsamples+1 ; i ++ )
		{
			sample[i] = 0 ;
		}
		if ( smooth_ == 1 )
		{
			weights[0] = 1.0;
			weights[1] = 1.0;
			weights[2] = 1.0;
			weights[3] = 1.0;
			weights[4] = 1.0;
			weights[5] = 0.8;
			weights[6] = 0.6;
			weights[7] = 0.4;
			weights[8] = 0.2;
			weights[9] = 0;
		}
		else
		{
			weights[0] = 1.0;
			weights[1] = 1.0;
			weights[2] = 1.0;
			weights[3] = 1.0;
			weights[4] = 0.8;
			weights[5] = 0.6;
			weights[6] = 0.4;
			weights[7] = 0.2;
			weights[8] = 0;
		}
		for ( int i = 0; i < numsamples+1; i ++ )
		{
			mult[i] = 1.0 ;
		}
	}

	/* this is the first packet from the joined smcc layer */
	if ( ( join_pending ) && ( layer_number == highest_smcc_layer ) )
	{

		prevmaxseq = maxseq[layer_number] =smcch->seqno -1;
		// initialize last_sample
		last_sample[layer_number] = smcch->seqno;
	}

	psize_ = smcch->psize;
	avgsize_ = ( int ) ( ( avgsize_ == 0 ) ? psize_ : ( avgsize_ * 0.9 + psize_ * 0.1 ) );
	prev_last_arrival_ = ( last_arrival_ == 0 ) ? now : last_arrival_; // no packet gap for first packet
	last_timestamp_ = smcch->timestamp;
	max_rtt = smcch->max_rtt;
	rate_change = smcch->rate_change;

	if ( smcch->is_clr )
	{
		if ( ( smcch->rtt_recv_id == recv_id ) )
		{
			representative = 1;
		}
		else if ( smcch->rtt_recv_id == -1 )
		{
			representative = 0;
			got_request_for_new_clr_on_this_layer = layer_number;
			new_round = 1;
			printf ( " @@ recv %d received new clr selection on L %d \n",recv_id,layer_number );
		}
		else
			representative = 0;
	}
	if ( smcch->clr_id[highest_smcc_layer] == recv_id )
		representative = 1;
	else
		representative = 0;


	if ( smcch->rtt_recv_id == recv_id )
	{
		rtt_cur = now - smcch->timestamp_echo - smcch->timestamp_offset;
		one_way_ = smcch->timestamp - smcch->timestamp_echo - smcch->timestamp_offset;
		if ( rtt_ > 0 )
		{
			double tmp_df_ = representative ? df_ : 0.5;
			rtt_est = rtt_ = tmp_df_ * rtt_ + ( 1-tmp_df_ ) * ( rtt_cur );
		}
		else
		{
			rtt_est = rtt_ = rtt_cur;

			// adjust previously faked lost history!
			if ( false_sample == -1 && sample_count <= numsamples+1 )
			{
				if ( rtt_ < rtt_lossinit )
				{
					sample[sample_count-1] = int ( sample[sample_count-1] *
					                               ( rtt_/rtt_lossinit ) * ( rtt_/rtt_lossinit ) );
					if ( sample[sample_count-1] < 1 )
						sample[sample_count-1] = 1;
				}
			}
		}
	}
	else
	{
		// do we have an initial measurement?
		if ( one_way_ > 0 )
		{
			rtt_cur = one_way_ + ( now - smcch->timestamp );
			double tmp_df_ =  0.9;
			rtt_est = rtt_cur =  tmp_df_ * rtt_ + ( 1-tmp_df_ ) * ( rtt_cur );
		}
	}
	add_packet_to_history ( pkt );

	if ( ( loss_seen_yet == 0 ) && ( smcch->seqno - prevmaxseq > 1 ) )
	{
		loss_seen_yet = 1;
		printf ( " at %f @@ first lost from receiver %d  @@@ \n",now,recv_id );
		if ( adjust_history_after_ss )
		{
			adjust_history ( smcch->timestamp,layer_number );
		}
	}


	if ( rtt_est > 0 )
		conservative_rtt = rtt_est;
	else
		conservative_rtt = max_rtt;

	p = est_loss ( layer_number );

	if ( ( p == 0 ) )
	{
		rate_ = 2 * est_thput ( layer_number );
	}
	else
	{
		rate_ = p_to_b ( p, conservative_rtt, 4 * conservative_rtt, fairsize_, bval_ );
	}

	Mrate = rate_*8/1000000;

	// report a constant rate (for debugging etc)
	if ( constant_rate_ > 0 )
		rate_ = constant_rate_ / 8.0;

	if ( layer_number == highest_smcc_layer )
	{
		suppression_rate = smcch->supp_rate;
	}

	int smcc_round_id;
	if ( layer_number == highest_smcc_layer )
		smcc_round_id = smcch->round_id;
	else
		smcc_round_id = 0;

	if ( round_id[layer_number] < smcch->round_id || round_id[layer_number] > smcch->round_id + MAX_ROUND_ID / 2 )
	{
		round_id[layer_number] = smcch->round_id;
		new_round = 1;
	}
	else
	{
		new_round = 0;
	}

	// use rate at beginning of round for suppression to avoid
	// possible feedback implosion in case of rate reduction during the round
	if ( new_round )
		comparison_rate = rate_;


	if ( ( rate_ < max_rate_layer[highest_smcc_layer] ) )
	{
		while ( highest_bcl_layer >= 0 )
		{
			printf ( " at %f rece %d, @@@ LEAVE  BCL  due to  rate is less than the min layer rate. cur rate %f @@\n",
			         now,recv_id,rate_*8/1000000 );
			Tcl::instance().evalf ( "%s leave %d", name(), bcl_addresses_smcc[highest_bcl_layer--] );
			bcl_join_pending = 0;
		}
		flag_join_bcl = 0;
	}

	receiver_leave = 0;

	if ( ( layer_number == highest_smcc_layer ) && ( rate_change == CONG_AVOID ) )
	{

		if ( ( receiving_rate[layer_number] ==  max_rate_layer[highest_smcc_layer] ) && ( join_next_layer == 1 ) &&
		        ( rate_>= receiving_rate[layer_number] ) && ( layer_number == highest_smcc_layer ) )
		{
			Tcl::instance().evalf ( "%s join %d", name(), smcc_addresses_smcc[++highest_smcc_layer] );
			join_pending = 1;
			leave_pending = 0;
			num_received_bcl_packet = 0;
			printf ( " at %f recv %d @@@ JOIN next smcc layer ,rate %f receiving rate %f join %d @@\n",now, recv_id, rate_*8/1000000,receiving_rate[layer_number]*8/1000000,join_next_layer );
			return;
		}

		if ( rate_ >= max_rate_layer[highest_smcc_layer]*damping_factor )
		{
			join_rate_ = rate_;
			if ( !loss_seen_yet )
				flag_join_bcl = 0;
			else if ( !representative )
				flag_join_bcl = get_join_time ( receiving_rate[layer_number], next_layer_rate );
			else if ( ( receiving_rate[highest_smcc_layer] >= max_rate_layer[highest_smcc_layer] )
			          && ( representative ) )
			{
				flag_join_bcl = get_join_time ( receiving_rate[layer_number], next_layer_rate );
			}
		}

	}

	if ( flag_join_bcl )
	{
		if ( ( ( round_num_bcl % flag_join_bcl ) == 0 ) && ( !bcl_join_pending ) && ( loss_seen_yet ) )
		{
//			printf(" at %f @@ R %d, try to join bcl layer rate %f round_num %d\n",
//				now, recv_id,join_rate_*8/1000000 ,round_num_bcl );
			last_seq_on_bcl_layer[0] = -2;
			join_bcl_layer ( 0 ); // Join bcl layer 0
			highest_bcl_layer = 0 ;
			bcl_join_pending = 1;
			num_received_bcl_packet = 0;
			round_num_when_join_bcl = round_num_bcl  ;
		}
	}


	if ( ( highest_smcc_layer != 0 ) && ( rate_ <= max_rate_layer[highest_smcc_layer -1 ] ) )
	{
		printf ( " @@@ reciver %d  leave highest %d\n",recv_id,highest_smcc_layer );
		printf ( "tmpr%i %f : l%f r%f -> %f (oo) R %i (new_R %i fb %f, add %f) | ",
		         recv_id, now, p, conservative_rtt, rate_ * 8.0 / 1000000.0,
		         round_id[layer_number], new_round, fb_time,
		         MAX ( 0.0, now - prev_last_arrival_ - max_rtt ) );
		for ( int i = 0 ; i <= numsamples; i++ )
			printf ( "%i ",sample[i] );
		printf ( " | %i %i\n", sample_count, numsamples );
		leave_highest_smcc_layer() ;
		fb_time = 0;
		receiver_leave = 0;
		Packet::free ( pkt );
		last_arrival_ = now;
		last_arrival_layer = layer_number;
		return;
	}

	if ( representative && tight_loop_ )
	{
		/* when the computed rate is less than the max rate of smcc */
		// feedback report once per RTT
		if ( ( now - last_report_sent > conservative_rtt ) )
		{
			fb_time = 0;
			sendpkt ( highest_smcc_layer );

			// consider the case when it joins the new layer from the sender's join request
			// it may received request new clr selection previous
			// cancel the feedback for new clr selection .
			if ( got_request_for_new_clr_on_this_layer == highest_smcc_layer )
				got_request_for_new_clr_on_this_layer = -1;

			// if it receives new clr selection request, reply it on its feedback
			if ( ( got_request_for_new_clr_on_this_layer >= 0 ) &&
			        ( got_request_for_new_clr_on_this_layer <= highest_smcc_layer ) )
			{
				sendpkt ( got_request_for_new_clr_on_this_layer );
				fb_time = 0;
				got_request_for_new_clr_on_this_layer = -1;
			}
		}
	}
	else   // not representative
	{
		if ( new_round )
		{
			double temp_time =  now + feedback_time ( max_rtt, MAX_NUM_RECVS );
			fb_time = temp_time;
		}

		if ( fb_time > 0 )
		{
			// reschedule packet in case last data packet was received too long ago
			// to prevent implosion (max. RTT should not be less than inter-packet gap)
			// (but don't reschedule if it's the first data packet)
			if ( now - prev_last_arrival_ > max_rtt )
				fb_time += now - prev_last_arrival_ - max_rtt;

			if ( comparison_rate >= suppression_rate || rate_ >= suppression_rate )
			{
				// suppress higher rate feedback
				fb_time = 0;
			}
			else
			{
				// send non-CLR feedback if it's not the first data packet and fb_time passed
				// (old feedback needs to be sent before checking for a new round!)
				if ( rate_ > 0 && fb_time <= now )
				{
					// send feedback if it receives request for new clr selection
					if ( ( got_request_for_new_clr_on_this_layer >= 0 ) &&
					        ( got_request_for_new_clr_on_this_layer <= highest_smcc_layer ) )
					{
						sendpkt ( got_request_for_new_clr_on_this_layer );
						fb_time = 0;
						got_request_for_new_clr_on_this_layer = -1;

					}
					else
						if ( ( highest_smcc_layer != 0 ) && ( rate_ <= max_rate_layer[highest_smcc_layer -1 ] ) )
						{
							receiver_leave = 0;
							fb_time = 0;
						}
						else
						{
							sendpkt ( highest_smcc_layer );
							fb_time = 0;
						}
				}
			}
		} //  end of if(fb_time>0)
	}
	if ( !printStatus_ )
	{
		if ( suppression_rate < DBL_MAX )
			printf ( "tmpr%i %f : L%d l%f r%f -> %f (%f) R %i (new_R %i fb %f, add %f) | ",
			         recv_id, now, highest_smcc_layer,p, conservative_rtt, rate_ * 8.0 / 1000000.0,
			         smcch->supp_rate * 8.0 / 1000000.0, round_id[layer_number], new_round, fb_time,
			         MAX ( 0.0, now - prev_last_arrival_ - max_rtt ) );
		else
			printf ( "tmpr%i %f : L%d l%f r%f -> %f (oo) R %i (new_R %i fb %f, add %f) | ",
			         recv_id, now, highest_smcc_layer,p, conservative_rtt, rate_ * 8.0 / 1000000.0,
			         round_id[layer_number], new_round, fb_time,
			         MAX ( 0.0, now - prev_last_arrival_ - max_rtt ) );
		if ( representative )
			printf ( " !! " );
		for ( int i = 0 ; i <= numsamples; i++ )
			printf ( "%i ",sample[i] );
		printf ( " | %i %i\n", sample_count, numsamples );
	}


	Packet::free ( pkt );
	last_arrival_layer = layer_number;
}
void SmccSinkAgent::bcl_recv ( Packet *pkt )
{
//	printf ( "SmccSinkAgent::bcl_recv(Packet *pkt\n" );
	hdr_smcc *smcch = hdr_smcc::access ( pkt );
	int round_num = smcch->round_num_bcl;
	double now = Scheduler::instance().clock();
	int join_next_bcl_layer;

	/* packet from leaved layer */
	if ( highest_bcl_layer == -1 )
		return;

	/* new packet on the joined layer */
	if ( last_seq_on_bcl_layer[smcch->bcl_layer] == -2 )
		last_seq_on_bcl_layer[smcch->bcl_layer] = smcch->seqno -1  ;

	if ( smcch->bcl_layer == highest_bcl_layer )
		join_next_bcl_layer = smcch->join_next_bcl;
	else
		join_next_bcl_layer = 0 ;


	if ( smcch->seqno != ( last_seq_on_bcl_layer[smcch->bcl_layer] +1 ) )
	{
		while ( highest_bcl_layer >= 0 )
		{
			printf ( " at %f rece %d, @@ LEAVE  BCL due to packet loss \n",now, recv_id );
			Tcl::instance().evalf ( "%s leave %d", name(), bcl_addresses_smcc[highest_bcl_layer--] );
		}
		num_received_bcl_packet = 0;
		flag_join_bcl = 0;
		bcl_join_pending = 0;
		maxseq[0]--; // to make a change on the loss history. It mimics a packet loss on the first smcc layer.
		last_sample[0]--; // to make a change on the loss history. It mimics a packet loss on the first smcc layer.
		printf ( " at %f @@@ recv_id %d BCL packet lost layer %d : reduce maxseq %d \n",now,recv_id,smcch->bcl_layer,maxseq[0] );
		return;
	}

	last_seq_on_bcl_layer[smcch->bcl_layer] = smcch->seqno ;


	int temp = round_num%flag_join_bcl;

	if ( highest_bcl_layer == 0 )
		/* condiser join delay */
		if ( temp >= ( pow ( 2,highest_bcl_layer +2 ) -2 ) )
		{
			join_bcl_layer ( ++highest_bcl_layer );
			last_seq_on_bcl_layer[highest_bcl_layer] == -2 ;
			printf ( " @@@ JOIN next bcl on L %d  Consider join delay\n",highest_bcl_layer );
		}


	if ( ( ( pow ( 2,highest_bcl_layer+2 ) -2 ) == temp ) && ( smcch->bcl_layer == 0 ) )
	{
		join_bcl_layer ( ++highest_bcl_layer );
		last_seq_on_bcl_layer[highest_bcl_layer] == -2 ;
		printf ( " at %f @@ JOIN next bcl %d\n",now,highest_bcl_layer );
	}
	num_received_bcl_packet++;


	if ( ( round_num - round_num_when_join_bcl ) >=  flag_join_bcl )
	{
		while ( highest_bcl_layer >= 0 )
		{
			printf ( " at %f recv_id %d @@ Leave BCL @@ \n",now, recv_id );
			Tcl::instance().evalf ( "%s leave %d", name(), bcl_addresses_smcc[highest_bcl_layer--] );
			bcl_join_pending = 0;
		}
		Tcl::instance().evalf ( "%s join %d", name(), smcc_addresses_smcc[++highest_smcc_layer] );
		join_pending = 1;
		leave_pending = 0;
		printf ( " at %f recv_id %d @@ JOIN NEXT SMCC %d @@@ \n",
		         now, recv_id,highest_smcc_layer );

// send message to the sender since there might be no receiver on next layer
		sendpkt ( highest_smcc_layer );
		receiver_leave = 0;
		flag_join_bcl = 0;
		bcl_join_time = now;
	}


}
void SmccSinkAgent::leave_highest_smcc_layer()
{
//	double now = Scheduler::instance().clock();
//	printf(" at %f rece %d, @@@ LEAVE  highest smcc layer %d @@@\n",now,
//			recv_id,highest_smcc_layer);
	Tcl::instance().evalf ( "%s leave %d", name(), smcc_addresses_smcc[highest_smcc_layer] );
	leave_pending = 1;
//	printf(" at %f @@@ receiver %d rate %f  max_rate %f \n", now,recv_id,
//		rate_*8/1000000,max_rate_layer[highest_smcc_layer-1]*8/1000000);
	fb_time = 0;
	if ( representative )
	{
		sendpkt ( highest_smcc_layer-1 );
		rate_ = max_rate_layer[highest_smcc_layer -1];
		receiver_leave = 1;
		sendpkt ( highest_smcc_layer );
	}
	highest_smcc_layer--;
	if ( bcl_join_pending )
		while ( highest_bcl_layer >= 0 )
		{
//		printf(" at %f rece %d, @@@ LEAVE  BCL @@@ \n",now,recv_id);
			Tcl::instance().evalf ( "%s leave %d", name(), bcl_addresses_smcc[highest_bcl_layer--] );
			bcl_join_pending = 0;
		}
}

void SmccSinkAgent::join_bcl_layer ( int bcl_layer_number )
{
	double now = Scheduler::instance().clock();
	Tcl::instance().evalf ( "%s join %d", name(), bcl_addresses_smcc[bcl_layer_number] );
	last_seq_on_bcl_layer[bcl_layer_number] = -2;
	bcl_join_pending = 1;
	printf ( " at %f @@@ recv %d JOIN BCL LAYER %d\n",now,recv_id,bcl_layer_number );
}

int SmccSinkAgent::get_join_time ( double rate, double next_rate )
{

//	printf(" recv %d current rate %f next rate %f\n",recv_id,rate*8/1000000,next_rate*8/1000000);
	double diff ;
	if ( next_rate == 0 )
		diff = max_rate_layer[highest_smcc_layer] - rate;
	else
		diff = next_rate - rate;

	expected_num_packet_from_bcl = ( int ) ( diff * min_rtt_smcc_ / avgsize_ ); // per RTT
	int i ;
	for ( i= 0; i < Max_smcc_Layer ; i++ )
	{
		if ( ( pow ( 2,i+2 )-2 ) >= expected_num_packet_from_bcl )
			break;
	}
	return ( ( int ) pow ( 2,i+1 ) );  // return the length of cycle of highest bcl layer for that num of packets

}

void SmccSinkAgent::add_packet_to_history ( Packet *pkt )
{
	hdr_smcc *smcch = hdr_smcc::access ( pkt );
	double now = Scheduler::instance().clock();
	register int i;
	register int seqno = smcch->seqno;
	int layer_number = smcch->smcc_layer;

	if ( lossvec_[layer_number] == NULL )
	{
		// Initializing history.
		rtvec_[layer_number]= ( double * ) malloc ( sizeof ( double ) *hsz );
		tsvec_[layer_number]= ( double * ) malloc ( sizeof ( double ) *hsz );
		lossvec_[layer_number]= ( char * ) malloc ( sizeof ( char ) *hsz );
		sizevec_[layer_number]= ( int * ) malloc ( sizeof ( int ) *hsz );
		last_recv_layer_vec_[layer_number]= ( int * ) malloc ( sizeof ( int ) *hsz );
		if ( rtvec_[layer_number] && lossvec_[layer_number] )
		{
			for ( i = 0; i <= maxseq[layer_number] ; i++ )
			{
				lossvec_[layer_number][i] = NOLOSS;
				rtvec_[layer_number][i] = now;
				tsvec_[layer_number][i] = last_timestamp_;
				sizevec_[layer_number][i] = fairsize_;
				if ( layer_number )
					last_recv_layer_vec_[layer_number][i] = layer_number -1;
				else
					last_recv_layer_vec_[layer_number][i] = 0;
			}
			for ( i = maxseq[layer_number]+1; i < hsz ; i ++ )
			{
				lossvec_[layer_number][i] = UNKNOWN;
				rtvec_[layer_number][i] = -1;
				tsvec_[layer_number][i] = -1;
				sizevec_[layer_number][i] = 0;
				last_recv_layer_vec_[layer_number][i] = 0;
			}
		}
		else
		{
			fprintf ( stderr, "error allocating memory for packet buffers\n" );
			abort ();
		}
	}

	if ( join_pending )
	{
		if ( layer_number == highest_smcc_layer )
		{
			for ( i = 0; i <= maxseq[layer_number] ; i ++ )
			{
				lossvec_[layer_number][i] = UNKNOWN;
				rtvec_[layer_number][i] = now;
				tsvec_[layer_number][i] = now;
				sizevec_[layer_number][i] = fairsize_;
			}
			join_pending = 0;
		}
	}

	if ( smcch->seqno - last_sample[layer_number] > hsz )
	{
		fprintf ( stderr, "time=%f, pkt=%d, last=%d history to small\n",
		          now, smcch->seqno, last_sample[layer_number] );
		abort();
	}


	/* for the time being, we will ignore out of order and duplicate
	   packets etc. */

	if ( seqno > maxseq[layer_number] )
	{
		rtvec_[layer_number][seqno%hsz]=now;
		tsvec_[layer_number][seqno%hsz]=last_timestamp_;
		lossvec_[layer_number][seqno%hsz] = RCVD;
		last_recv_layer_vec_[layer_number][seqno%hsz] = last_arrival_layer;
		sizevec_[layer_number][seqno%hsz] = smcch->psize;
		i = maxseq[layer_number]+1;
		if ( i < seqno )
		{
			double delta = ( tsvec_[layer_number][seqno%hsz]-
			                 tsvec_[layer_number][maxseq[layer_number]%hsz] ) / ( seqno-maxseq[layer_number] ) ;
			double tstamp = tsvec_[layer_number][maxseq[layer_number]%hsz]+delta ;
			while ( i < seqno )
			{
				rtvec_[layer_number][i%hsz]=now;
				tsvec_[layer_number][i%hsz]=tstamp;
				sizevec_[layer_number][i%hsz] = avgsize_;

				if ( tsvec_[layer_number][i%hsz]-lastloss > conservative_rtt )
				{
					// Lost packets are marked as "LOST"
					// at most once per RTT.
					lossvec_[layer_number][i%hsz] = LOST;
					printf ( " at %f @@@##### LOST #### on reveiver: %d , seq %d\n",now,recv_id,i );
					lastloss = tstamp;
					// GK
					total_pkts_lost_++;
					if ( bcl_join_pending )
						while ( highest_bcl_layer >= 0 )
						{
							printf ( " at %f rece %d, @@@ LEAVE  BCL  During Join Attempt Due to Packet Loss @@@\n", now,recv_id );
							Tcl::instance().evalf ( "%s leave %d", name(),
							                        bcl_addresses_smcc[highest_bcl_layer--] );
							bcl_join_pending = 0;
						}
				}
				else
				{
					// This lost packet is marked "NOLOSS"
					// because it does not begin a loss event.
					lossvec_[layer_number][i%hsz] = NOLOSS;
				}
				i++;
				tstamp = tstamp+delta;
			}
		}
		maxseq[layer_number] = seqno;
	}
}

/*
 * Estimate the loss rate.  This function calculates two loss rates,
 * and returns the smaller of the two.
 */
double SmccSinkAgent::est_loss ( int layer_number )
{
	int i;
	double ave_interval1, ave_interval2;
//	double now = Scheduler::instance().clock();
	int ds ;

	// sample[i] counts the number of packets since the i-th loss event
	// sample[0] contains the most recent sample.

	//packets (and possibly loss events) are aggregated to form fairsize_'ed packets

	for ( i = last_sample[layer_number]; i <= maxseq[layer_number] ; i ++ )
	{
		if ( aggregate_packets_ )
		{
			if ( ( !aggregate_loss_ && lossvec_[layer_number][i%hsz] == LOST ) )
			{
				// adjust sample if each loss event counts as full packet
				sum_psize_ += fairsize_;
			}
			else
			{
				sum_psize_ += sizevec_[layer_number][i%hsz];
			}

			if ( sum_psize_ >= 0 )
			{
				// after received packet ignore the next packets until fairsize_ bytes were received
				sum_psize_ -= fairsize_;
				sample[0]++;
			}
		}
		else
		{
			sample[0]++;
		}

		if ( lossvec_[layer_number][i%hsz] == NOLOSS || lossvec_[layer_number][i%hsz] == LOST )
			sum_lsize_ += sizevec_[layer_number][i%hsz];

		if ( ( aggregate_loss_ && sum_lsize_ >= 0 ) ||
		        ( !aggregate_loss_ && lossvec_[layer_number][i%hsz] == LOST ) )
		{
			sum_lsize_ -= fairsize_;
			//  new loss event
			sample_count++;
			shift_array ( sample, numsamples+1, 0 );
			multiply_array ( mult, numsamples+1, mult_factor_ );
			shift_array ( mult, numsamples+1, 1.0 );
			mult_factor_ = 1.0;
		}
	}
	last_sample[layer_number] = maxseq[layer_number]+1 ;
	/* consider number of received packet from bcl */
	sample[0] += num_received_bcl_packet;
	num_received_bcl_packet = 0;

	if ( sample_count>numsamples+1 )
		// The array of loss intervals is full.
		ds=numsamples+1;
	else
		ds=sample_count;

	if ( sample_count == 1 && false_sample == 0 )
		// no losses yet
		return 0;

	/* do we need to discount weights? */
	if ( sample_count > 1 && discount && sample[0] > 0 )
	{
		double ave = weighted_average ( 1, ds, 1.0, mult, weights, sample );
		int factor = 2;
		double ratio = ( factor*ave ) /sample[0];
		double min_ratio = 0.5;
		if ( ratio < 1.0 )
		{
			// the most recent loss interval is very large
			mult_factor_ = ratio;
			if ( mult_factor_ < min_ratio )
				mult_factor_ = min_ratio;
		}
	}
	// Calculations including the most recent loss interval.
	ave_interval1 = weighted_average ( 0, ds, mult_factor_, mult, weights, sample );
	// The most recent loss interval does not end in a loss
	// event.  Include the most recent interval in the
	// calculations only if this increases the estimated loss
	// interval.
	ave_interval2 = weighted_average ( 1, ds, mult_factor_, mult, weights, sample );
	if ( ave_interval2 > ave_interval1 )
		ave_interval1 = ave_interval2;
	if ( ave_interval1 >= 0.99 )
	{
		//print_loss(sample[0], ave_interval1);
		return 1/ave_interval1;
	}
	else
	{
		fprintf ( stderr, "r%i est_loss(): sample_count %i, false_sample %i\n", recv_id, sample_count , false_sample );
		double avg = 0;
		for ( int i = 0 ; i <= numsamples; i++ )
		{
			fprintf ( stderr, "%i ", sample[i] );
			avg += sample[i];
		}
		fprintf ( stderr, " --- avg %f %i\n", ave_interval1, numsamples );


		exit ( 1 );
		//return 999;
	}
}

void SmccSinkAgent::print_loss ( int sample, double ave_interval )
{
//	double now = Scheduler::instance().clock();
	double drops = 1/ave_interval;
}

// Calculate the weighted average.
double SmccSinkAgent::weighted_average ( int start, int end, double factor, double *m, double *w, int *sample )
{
	int i;
	double wsum = 0;
	double answer = 0;
	if ( smooth_ == 1 && start == 0 )
	{
		if ( end == numsamples+1 )
		{
			// the array is full, but we don't want to uses
			//  the last loss interval in the array
			end = end-1;
		}
		// effectively shift the weight arrays
		for ( i = start ; i < end; i++ )
			if ( i==0 )
				wsum += m[i]*w[i+1];
			else
				wsum += factor*m[i]*w[i+1];
		for ( i = start ; i < end; i++ )
			if ( i==0 )
				answer += m[i]*w[i+1]*sample[i]/wsum;
			else
				answer += factor*m[i]*w[i+1]*sample[i]/wsum;
		return answer;
	}
	else
	{
		for ( i = start ; i < end; i++ )
			if ( i==0 )
				wsum += m[i]*w[i];
			else
				wsum += factor*m[i]*w[i];
		for ( i = start ; i < end; i++ )
			if ( i==0 )
				answer += m[i]*w[i]*sample[i]/wsum;
			else
				answer += factor*m[i]*w[i]*sample[i]/wsum;
		return answer;
	}
}

// Shift array a[] up, starting with a[sz-2] -> a[sz-1].
void SmccSinkAgent::shift_array ( int *a, int sz, int defval )
{
	int i ;
	for ( i = sz-2 ; i >= 0 ; i-- )
	{
		a[i+1] = a[i] ;
	}
	a[0] = defval;
}

void SmccSinkAgent::shift_array ( double *a, int sz, double defval )
{
	int i ;
	for ( i = sz-2 ; i >= 0 ; i-- )
	{
		a[i+1] = a[i] ;
	}
	a[0] = defval;
}

// Multiply array by value, starting with array index 1.
// Array index 0 of the unshifted array contains the most recent interval.
void SmccSinkAgent::multiply_array ( double *a, int sz, double multiplier )
{
	int i ;
	for ( i = 1; i <= sz-1; i++ )
	{
		double old = a[i];
		a[i] = old * multiplier ;
	}
}

/*
 * compute estimated throughput for report.
 */
double SmccSinkAgent::est_thput ( int layer_number )
{
// return Byte/s
	double last = rtvec_[layer_number][maxseq[layer_number]%hsz];
//	double now = Scheduler::instance().clock();
	int rcvd = 0;
	int l = highest_smcc_layer;
	int m[highest_smcc_layer+1];
//	int layer_index = layer_number;
	int j = layer_number;
	double last_recent_time = last_arrival_;
	while ( l >= 0 )
	{
		m[l] = maxseq[l];
		l--;
	}

	// count number of packets received in the last 2 RTTs (min. of 10 packets)
	while ( m[j] > 0 && ( rcvd < 10 || ( last - last_recent_time < 2*conservative_rtt ) ) )
	{
		if ( lossvec_[j][m[j]%hsz] == RCVD )
		{
			rcvd++;
		}
		j = last_recv_layer_vec_[j][m[j]--%hsz];
		last_recent_time =  rtvec_[j][m[j]%hsz];

	}
	if ( ( last - last_recent_time ) > 0 )
	{
		return rcvd/ ( last - last_recent_time ) * avgsize_;
	}
	else
	{
		return 0;
	}
}

double SmccSinkAgent::adjust_history ( double ts,int layer_number )
{
	int i;
	double p;
	for ( i = maxseq[layer_number]; i >= 0 ; i-- )
	{
		if ( lossvec_[layer_number][i%hsz] == LOST )
		{
			lossvec_[layer_number][i%hsz] = NOLOSS;
		}
	}
	lastloss = ts;
	// ignore previous loss history
	last_sample[layer_number] = maxseq[layer_number] + 1;

	// no need to divide throughput by 2 since its actual receive rate, not sending rate
	p=b_to_p ( est_thput ( layer_number ), conservative_rtt, 4 * conservative_rtt, fairsize_, bval_ );
	false_sample = ( int ) ( 1.0/p + 0.5 );
	sample[1] = false_sample;
	sample[0] = 0;
	sample_count = 2;
	false_sample = -1 ;
	rtt_lossinit = conservative_rtt;
	return p;
}


/*
 * Unused
 */
void SmccSinkAgent::nextpkt()
{
	sendpkt ( highest_smcc_layer );
}

/*
 * Create report message, and send it.
 */
void SmccSinkAgent::sendpkt ( int layer_number )
{
	double now = Scheduler::instance().clock();

	if ( !active_ ) return;

	Packet* pkt = allocpkt();
	if ( pkt == NULL )
	{
		fprintf ( stderr, "error allocating packet\n" );
		abort();
	}

	hdr_smcc_ack *smcc_ackh = hdr_smcc_ack::access ( pkt );

	smcc_ackh->recv_id = recv_id;
	smcc_ackh->smcc_layer = layer_number;
	smcc_ackh->timestamp = now;
	smcc_ackh->timestamp_echo = last_timestamp_;
	smcc_ackh->timestamp_offset = now - last_arrival_;

	
	
	if ( ( got_request_for_new_clr_on_this_layer >= 0 ) && ( highest_smcc_layer > layer_number ) &&
	        ( got_request_for_new_clr_on_this_layer <= highest_smcc_layer ) && ( !receiver_leave ) )
	{
		smcc_ackh->smcc_layer = got_request_for_new_clr_on_this_layer;
		smcc_ackh->rate = max_rate_layer[got_request_for_new_clr_on_this_layer];
		printf ( " @@## rec %d send feedback on L %d rate %f got %d\n",
		         recv_id, layer_number,smcc_ackh->rate*8/1000000,got_request_for_new_clr_on_this_layer );
	}
	else
		smcc_ackh->rate = rate_;

	if ( ( layer_number ) && ( rate_ < max_rate_layer[layer_number -1 ] ) && rate_ )
	{
		printf ( " recv %d at %f @@@ rate is less than lower layer,rate %f \n",recv_id,now,rate_*8/1000000 );
		exit ( 0 );
	}

	smcc_ackh->round_id = round_id[layer_number];
	smcc_ackh->have_rtt = ( rtt_ != 0 );
	smcc_ackh->have_loss = loss_seen_yet;
	smcc_ackh->receiver_leave = receiver_leave;
	last_report_sent = now;
	rcvd_since_last_report = 0;
	send ( pkt, 0 );
	

	if ( printStatus_ )
	{
		if ( suppression_rate < DBL_MAX )
			printf ( "r%i L %d %f : %f %f -> %2.1f (supp %1f(%1f) fb %f round %i max_rtt %f r%il%i)%c [",
			         smcc_ackh->recv_id, layer_number, now, est_loss ( layer_number ), rtt_,
			         smcc_ackh->rate * 8.0 / 1000000.0,
			         suppression_rate * 8.0 / 1000000.0, comparison_rate * 8.0 / 1000000.0,
			         fb_time, round_id[layer_number], max_rtt, smcc_ackh->have_rtt, smcc_ackh->have_loss,
			         representative ? '!' : ' ' );
		else
			printf ( "r%i L %d %f : %f %f -> %f (supp oo(%1f) fb %f round %i max_rtt %f r%il%i)%c [",
			         smcc_ackh->recv_id, layer_number ,now, est_loss ( layer_number ), rtt_,
			         smcc_ackh->rate * 8.0 / 1000000.0,
			         comparison_rate * 8.0 / 1000000.0,
			         fb_time, round_id[layer_number], max_rtt, smcc_ackh->have_rtt, smcc_ackh->have_loss,
			         representative ? '!' : ' ' );
		for ( int i = 0 ; i <= numsamples; i++ )
			printf ( "%i ", sample[i] );
		printf ( "]\n" );
	}

	/*
	printf("%f %i |\t",now,maxseq);
	double avg = 0;
	for (int i = 0 ; i <= numsamples; i++) {
	        printf("%i ",sample[i]);
	        avg += sample[i];
	}
	printf(" avg %f loss %f XXX\n", avg / (numsamples + 1), smcc_ackh->loss);
	*/
}

double SmccSinkAgent::feedback_time ( double rtt, int num_recv )
{
	double T = fbtime_mult_ * rtt;
	double x = Random::uniform();
	double t;

	t = T * ( 1.0 + log ( x ) / log ( ( double ) num_recv ) );
	if ( t < 0 ) t = 0;
	return t;
}

int SmccSinkAgent::command ( int argc, const char*const* argv )
{
	if ( argc==2 )
	{
		if ( strcmp ( argv[1],"begin" ) ==0 )
		{
			begin();
			return TCL_OK;
		}
		if ( strcmp ( argv[1],"test" ) ==0 )
		{
			test();
			return TCL_OK;
		}
		if ( strcmp ( argv[1],"stop" ) ==0 )
		{
			stop();
			return TCL_OK;
		}
	}
	else if ( argc == 3 )
	{
		if ( strcmp ( argv[1], "weights" ) == 0 )
		{
			/*
			 * weights is a string of numbers, seperated by + signs
			 * the first number is the total number of weights.
			 * the rest of them are the actual weights
			 * this overrides the defaults
			 */
			char *w ;
			w = ( char * ) calloc ( strlen ( argv[2] ) +1, sizeof ( char ) ) ;
			if ( w == NULL )
			{
				fprintf ( stderr, "error allocating w\n" );
				abort();
			}
			strcpy ( w, ( char * ) argv[2] );
			numsamples = atoi ( strtok ( w,"+" ) );
			sample = ( int * ) malloc ( ( numsamples+1 ) *sizeof ( int ) );
			weights = ( double * ) malloc ( ( numsamples+1 ) *sizeof ( double ) );
			mult = ( double * ) malloc ( ( numsamples+1 ) *sizeof ( double ) );
			fflush ( stdout );
			if ( sample && weights && mult )
			{
				int count = 0 ;
				while ( count < numsamples )
				{
					sample[count] = 0;
					mult[count] = 1;
					char *w;
					w = strtok ( NULL, "+" );
					if ( w == NULL )
						break ;
					else
					{
						weights[count++] = atof ( w );
					}
				}
				if ( count < numsamples )
				{
					fprintf ( stderr, "error in weights string %s\n", argv[2] );
					abort();
				}
				sample[count] = 0;
				weights[count] = 0;
				mult[count] = 1;
				free ( w );
				return ( TCL_OK );
			}
			else
			{
				fprintf ( stderr, "error allocating memory for smaple and weights:2\n" );
				abort();
			}
		}
	}
	return ( Agent::command ( argc, argv ) );
}

void SmccSinkAgent::stop()
{
//        rate_ = 0; // -> receiver leaves session
	active_ = 0;
	receiver_leave = 1;
	if ( representative )
	{
		if ( printStatus_ )
			printf ( "r%i %f : CLR leave-group\n", recv_id, Scheduler::instance().clock() );

	}
	while ( highest_smcc_layer >= 0 )
	{
		Tcl::instance().evalf ( "%s leave %d", name(), smcc_addresses_smcc[highest_smcc_layer] );
		receiver_leave = 1;
		sendpkt ( highest_smcc_layer-- );
	}
	
	print_statistics();
}

void SmccSinkAgent::print_statistics()
{
	printf("###### STATISTICS SMCC(%d) ######\n",recv_id );
	int total_pkts_sent_ = total_pkts_recv_ + total_pkts_lost_;
	double loss_ratio = (double)total_pkts_lost_ / (double)total_pkts_sent_;
	printf(" PACKETS SENT %d PACKETS RECEIVED %d PACKETS LOST %d LOSS RATIO %f \n",total_pkts_sent_,total_pkts_recv_, total_pkts_lost_, loss_ratio);

	
}
void SmccSinkAgent::begin() {
	Tcl & tcl = Tcl::instance();
	for ( int i=0; i < Max_smcc_Layer ; i++) {
                max_rate_layer[i] = base_layer_rate*pow(2,i) ; 
                assert(max_rate_layer[i] > DBL_MAX);
        }
	tcl.evalf("%s join  %d", name(), smcc_addresses_smcc[0] );
	printf(" receiver %d joined \n",recv_id);
}


void SmccSinkAgent::test()
{
	for ( int i=0; i < Max_smcc_Layer ; i++ )
	{
		max_rate_layer[i] = base_layer_rate*pow ( 2,i ) ;
		assert ( max_rate_layer[i] > DBL_MAX );
	}
	flag_join_bcl = get_join_time ( 0, max_rate_layer[3] );
	join_bcl_layer ( 0 );
	highest_bcl_layer++;
}

void SmccSinkAgent::bcl_recv_test ( Packet *pkt )
{
	hdr_smcc *smcch = hdr_smcc::access ( pkt );
//	int round_num = smcch->round_num_bcl;
	double now = Scheduler::instance().clock();
	int join_next_bcl_layer;


	/* new packet on the joined layer */
	if ( last_seq_on_bcl_layer[smcch->bcl_layer] == -2 )
		last_seq_on_bcl_layer[smcch->bcl_layer] = smcch->seqno -1  ;

	if ( smcch->bcl_layer == highest_bcl_layer )
		join_next_bcl_layer = smcch->join_next_bcl;
	else
		join_next_bcl_layer = 0 ;

	last_seq_on_bcl_layer[smcch->bcl_layer] = smcch->seqno ;

//	int temp = round_num%flag_join_bcl;

	if ( join_next_bcl_layer )
	{
		join_bcl_layer ( ++highest_bcl_layer );
		last_seq_on_bcl_layer[highest_bcl_layer] == -2 ;
		printf ( " at %f @@@ JOIN next bcl %d\n",now,highest_bcl_layer );
	}
	num_received_bcl_packet++;
}

