
// Min rtt for bcl layer 0 = 64ms

#include "agent.h"
#include "packet.h"
#include "ip.h"
#include "timer-handler.h"
#include "random.h"

#define SMALLFLOAT 0.0000001

/* modes of rate change */
#define INIT_RATE 0
#define SLOW_START 1
#define CONG_AVOID 2
#define BCL_Type 0
#define SMCC_Type 1

/* cut sending rate after DEC_NO_REPORT rtts without receiver report */
#define DEC_NO_REPORT 10
#define G_FACTOR 0.1
#define MAX_ROUND_ID 1000000 // just for logging, in real world 16 is enough
#define MAX_RTT_DECAY 0.9
#define Max_Bcl_Layer 6
#define Max_smcc_Layer 6 
#define min_rtt_smcc_ 0.064  // Min rtt for bcl layer 0 = 64ms
#define rate_smcc_first_layer 10000000  // BYTE/S  

#define LARGE_DOUBLE 9999999999.99 
#define SAMLLFLOAT 0.0000001

/* packet status */
#define UNKNOWN 0
#define RCVD 1
#define LOST 2
#define NOLOSS 3

#define DEFAULT_NUMSAMPLES 8 
#define MAX_NUM_RECVS 10000

enum {PRIO_FORCE = 4, PRIO_NO_RTT = 3, PRIO_RECV = 2, PRIO_CLR = 1};


struct hdr_smcc {
        int packet_type ; //  0 Bcl 1:TFMCC
        int layer_number ; //  
        int round_num_bcl;  // round number of rtt in the bcl 0 layer     
	int smcc_layer; // 
	int bcl_layer; // 

	int seqno;		 // data sequence number
	double supp_rate;	 // suppression rate
	double sending_rate;	 // suppression rate
	double next_layer_rate;	 // next layer rate  
	double next_smcc_layer_rate;	 // 
	double timestamp; 	 // time this message was sent
	double timestamp_echo; 	 // echo timestamp from last receiver report
	double timestamp_offset; // offset since sender received report
	double max_rtt;          // maximum RTT of all receivers
	int psize;		 // packet size
	int rate_change;	 // 
	int round_id; 		 // round id.
	int is_clr;              // current representative receiver
	int rtt_recv_id;         // id of receiver whose timestamp is sent back
	int join_next_layer;         //  
	int join_next_bcl;         //  
	int clr_id[Max_smcc_Layer];    //  

	static int offset_;	//offset for this header
	inline static int& offset() { return offset_; }
	inline static hdr_smcc* access(Packet* p) {
		return (hdr_smcc*) p->access(offset_);
	}
};

struct hdr_smcc_ack {
	int recv_id;                    // ID of receiver giving feedback
	double timestamp;		// time this report was sent
	double timestamp_echo;   	// echo timestamp from last data packet
	double timestamp_offset;        // offset since receiver got last packet
	double rate;                    // expected sending rate (eqn)
	int round_id; 		        // round id echo
	int have_rtt;                   // receiver has at least one RTT measurement
	int have_loss;                  // receiver exprienced at least one loss event
	int receiver_leave;             // receiver will leave session
	int smcc_layer; 		// 
	
	static int offset_;             // offset for this header
	inline static int& offset() 
		{ return offset_; }
	inline static hdr_smcc_ack* access(Packet* p) {
		return (hdr_smcc_ack*) p->access(offset_);
	}
};

class SmccAgent; 

class BclTimer : public TimerHandler {
public:
                BclTimer(SmccAgent *a) : TimerHandler() { a_ = a; }
                virtual void expire(Event *e);
                int counter_;
                double interval_;
                int layer_number_;
                int start_phase_;
		int join_next_bcl;         //  
protected:
                SmccAgent *a_;
};
// ******************************************


class SmccSendTimer : public TimerHandler {
public:
		SmccSendTimer(SmccAgent *a) : TimerHandler() { a_ = a; }
		virtual void expire(Event *e);
		int smcc_layer; // 
protected:
		SmccAgent *a_;
};	

class SmccNoFeedbackTimer : public TimerHandler {
public:
		SmccNoFeedbackTimer(SmccAgent *a) : TimerHandler() { a_ = a; }
		virtual void expire(Event *e);
		int smcc_layer; // 
protected:
		SmccAgent *a_;
}; 

class SmccAgent : public Agent {
//	friend SmccSendTimer;
//	friend SmccNoFeedbackTimer;
public:
	SmccAgent();
        void bcl_timeout(int,int); // 
       	void bcl_sendit(int,int); // 
      	void get_bcl_interval(int); // 
        void set_bcl_start(); // 
        void switch_layer(int,int); // 

	void recv(Packet*, Handler*);
	void sendpkt(int);
	void nextpkt(int);
	int command(int argc, const char*const* argv);
	void create_new_smcc_layer(int);
	void send_pkt_at_max_rate_smcc_layer(int);
	void start();
	void warmup();
	void stop(int);
	void increase_rate(int);
	void decrease_rate(int);
	void slowstart(int);
	void reduce_rate_on_no_feedback(int);
	void reduce_rate_to_zero_on_no_feedback(int);

protected:
	SmccSendTimer *send_timer_[Max_smcc_Layer];
	SmccNoFeedbackTimer *NoFeedbacktimer_[Max_smcc_Layer];
        BclTimer *bcl_timer_[Max_Bcl_Layer];  // 
        int seq_number_bcl[Max_Bcl_Layer]; // 
        int rtt_round_seqno;  // round number of rtt in the bcl 0 layer    
        int round_num_bcl;  // round number of rtt in the bcl 0 layer     

	int seqno_[Max_smcc_Layer] ;// 
	int join_next_layer[Max_smcc_Layer];         //  
	int bcl_round_seqno;         //  
	double next_sched_time;  // 
	int psize_;             // packet size of TFMCC flow
	int fairsize_;          // packet size of competing flow
	double rate_[Max_smcc_Layer];		// send rate 
	double max_rate_on_layer[Max_smcc_Layer];	// Max rate on each layer 
	double oldrate_[Max_smcc_Layer];	// allows rate to be changed gradually
	double delta_;		// allows rate to be changed gradually
	int rate_change_; 	// slow start, cong avoid, decrease ...
	double expected_rate[Max_smcc_Layer]; 	// TCP friendly rate based on current RTT 
				//  and recever-provded loss estimate
	double maxrate_[Max_smcc_Layer];	// maximum rate during slowstart (prevents
	                        //  sending at more than 2 times the 
				//  rate at which the receiver is _receiving_)
 	int printStatus_;	// to print status reports
	int sndr_id;            // only needed for status reports

	double inter_packet[Max_smcc_Layer];    // inter packet gap

	double max_rtt_[Max_smcc_Layer];        // (assumed) maximum RTT of all receivers
	double Initial_RTT_;    // safe initial value for max RTT
	double t_factor_;       // T = t_factor_ * max_rtt_
	double round_begin_[Max_smcc_Layer];    // time when fb round started

	int InitRate_;		// initial send rate
	double last_change_[Max_smcc_Layer];	// time last change in rate was made
	double ssmult_;		// during slow start, increase rate by this 
				//  factor every rtt
	int bval_;		// value of B for the formula
	double overhead_;	// if > 0, dither outgoing packets 
	TracedInt ndatapack_;	// number of packets sent
	int active_;		// have we shut down? 
	int round_id[Max_smcc_Layer];		// round id  
	int highest_slow_start_smcc_layer;		// round id  
	int base_layer_rate;

	double min_rate_;
	double supp_rate_[Max_smcc_Layer];

	// extensions for multicast
	int    rtt_recv_id[Max_smcc_Layer];
	int    rtt_prio[Max_smcc_Layer];
	double rtt_recv_timestamp[Max_smcc_Layer];
	double rtt_recv_last_feedback[Max_smcc_Layer];
	double rtt_rate[Max_smcc_Layer];
	double next_send_time;

	// CLR
	int    clr_id[Max_smcc_Layer];  //  
	double clr_timestamp[Max_smcc_Layer]; //   
	double clr_last_feedback[Max_smcc_Layer]; //    
};





class SmccSinkAgent : public Agent {
public:
	SmccSinkAgent();
	void recv(Packet*, Handler*);
	int command(int argc, const char*const* argv);
	void stop();
	void begin();
	void test();

protected:
	void sendpkt(int);
	void nextpkt();
	void increase_pvec(int);
	void add_packet_to_history(Packet *);
	void bcl_recv(Packet *); // 
	void bcl_recv_test(Packet *); // 
	double adjust_history(double,int);
	double est_loss(int);
	double est_thput(int); 
	void shift_array(int *a, int sz, int defval) ;
	void shift_array(double *a, int sz, double defval) ;
	void multiply_array(double *a, int sz, double multiplier);
	void print_loss(int sample, double ave_interval);
	double weighted_average(int start, int end, double factor, double *m, double *w, int *sample);
	double feedback_time(double rtt, int num_recv);
	void join_bcl_layer(int); // 
	void leave_highest_smcc_layer(); // 
	int get_join_time(double,double); // 

	int psize_;		// size of received packet
	int fairsize_;          // packet size of competing flow
	int avgsize_;           // avg size of sent packets (EWMA)
	int sum_psize_;
	int sum_lsize_;
	int aggregate_packets_; // bool
	int aggregate_loss_;    // bool
	double rtt_;		// smoothed measured rtt
	double rtt_est; 	// use adjusted one-way delay to continue to modify the rtt
	double rtt_lossinit;    // RTT when history was initialized
	double max_rtt;         // max RTT (from sender)
        double conservative_rtt;
	double one_way_;        // one-way delay from receiver to sender
	int smooth_;		// for the smoother method for incorporating
				//  incorporating new loss intervals

	// these assist in keep track of incming packets and calculate flost_
	double last_timestamp_, last_arrival_;
	int hsz;
	char *lossvec_[Max_smcc_Layer];
	int  *sizevec_[Max_smcc_Layer];
	int  *last_recv_layer_vec_[Max_smcc_Layer];
	double *rtvec_[Max_smcc_Layer];
	double *tsvec_[Max_smcc_Layer];
	int round_id[Max_smcc_Layer]  ; // 
	int highest_smcc_layer; // 
	int display_rate_; // 
	double damping_factor; // 
	int total_bytes_; // 
	int without_join_attempt; // 
	int last_arrival_layer; // 
	int highest_bcl_layer; // 
	int round_num_when_join_bcl; // 
	int num_received_bcl_packet; // 
	int expected_num_packet_from_bcl; // 
	int bcl_join_pending; // 
	int flag_join_bcl; // 
	int join_pending; // 
	int leave_pending; // 
	int last_seq_on_layer[Max_smcc_Layer]; // 
	int last_seq_on_bcl_layer[Max_smcc_Layer]; // 
	int seq_all; // 
	int layer_number; // 
	int round_num_bcl; // 
        int got_request_for_new_clr_on_this_layer; // 

	double max_rate_layer[Max_smcc_Layer]; // Acumulative rate
	int numsamples ;
	int *sample;
	double *weights ;
	double *mult ;
	double mult_factor_;	// most recent multiple of mult array
	int sample_count ;
	int last_sample[Max_smcc_Layer] ;  
	int maxseq[Max_smcc_Layer]; 		// max seq number seen
	int total_received_;	// total # of pkts rcvd by rcvr
	int bval_;		// value of B used in the formula
	double last_report_sent; 	// when was last feedback sent
	int rcvd_since_last_report; 	// # of packets rcvd since last report
	double lastloss; 	// when last loss occured
	int printStatus_;	// to print estimated loss rates
	int tight_loop_;	// flag for tight feedback loop to rep recv
	double fbtime_mult_;         // T = mult * RTT
	double fb_time;         // delay for feedback suppression
	double constant_rate_;  // set to positive value to use instead of calculated rate (Bit/s)
	double comparison_rate; // rate at beginning of feedback round (for suppression)
	double suppression_rate; // sender's suppression rate
	int receiver_leave;     // flag to indicate that CLR will leave group


	// these are for "faking" history after slow start
	int loss_seen_yet; 	// have we seen the first loss yet?
	int adjust_history_after_ss; // fake history after slow start? (0/1)
	int false_sample; 	// by how much?

	int discount ;		// emphasize most recent loss interval
				//  when it is very large
	int representative;     // flag: 1 = feedback receiver
	int recv_id;
	//int ca_; //Enable Sqrt(RTT) based congestion avoidance mode
	double df_;		// decay factor for accurate RTT estimate
        double rate_;
        double join_rate_;
        double receiving_rate[Max_smcc_Layer];
        double next_layer_rate;
        double bcl_join_time; // 
        int active_;
	int rate_change; // 
	int base_layer_rate;

	// for statistics
	int total_pkts_recv_;
	int total_pkts_lost_;
	void print_statistics();
	double end_2_end_delay_;
	double jitter_;
	double ptransit_;
}; 
