#
# Copyright (c) 2007 University of Patras Computer Engineering and Informatics Department Greece
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the MASH Research
# 	Group at the University of California Berkeley.
# 4. Neither the name of the University nor of the Research Group may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#


proc mvar args {
	upvar self _s
	uplevel $_s instvar $args
}

SimStream/RTP set uniq_srcid 0
SimStream/RTP proc alloc_srcid {} {
	set id [SimStream/RTP set uniq_srcid]
	SimStream/RTP set uniq_srcid [expr $id+1]
	return $id
}



SimStream/RTP instproc init {} {
	$self next
	mvar Stream1_ Stream2_ Stream3_  
	
	set Stream1_ [new Session/RTP] 
	set Stream2_ [new Session/RTP]
	set Stream3_ [new Session/RTP]
	
	
	set srcid_ [SimStream/RTP alloc_srcid]
	$self stream-id $srcid_
#	$self set streamid_  $streamid_

	$self 	  session $Stream1_ 
	$self 	  session $Stream2_
	$self 	  session $Stream3_
	$Stream1_ stream $self
	$Stream2_ stream $self
	$Stream3_ stream $self
	$self set stopped_ 1
}

SimStream/RTP instproc assign-groups { g1 g2 g3 } {
	mvar group1_ group2_ group3_ 
#	set g1 [expr $g1]
#	set g2 [expr $g2]
#	set g3 [expr $g3]
	$self set group1_ $g1
	$self set group2_ $g2
	$self set group3_ $g3
#	puts "Client <-----------------------------------"
#	puts [$self set group1_]
#	puts [$self set group2_]
#	puts [$self set group3_]
#	puts "Client ----------------------------------->"
}


SimStream/RTP instproc start {} {
	mvar group1_  
	if ![info exists group1_] {
		puts "error: can't transmit before joining group!"
		exit 1
	}
	mvar group2_
	if ![info exists group2_] {
		puts "error: can't transmit before joining group!"
		exit 1
	}
	mvar group3_
	if ![info exists group3_] {
		puts "error: can't transmit before joining group!"
		exit 1
	}
	mvar Stream1_  Stream2_  Stream3_
	$Stream1_ start 
	$Stream2_ start
	$Stream3_ start  
}

SimStream/RTP instproc report-interval { i } {
	mvar Stream1_  Stream2_  Stream3_
	$Stream1_ report-interval $i 
	$Stream2_ report-interval $i 
	$Stream3_ report-interval $i 
	
}


SimStream/RTP instproc start-session { g } {
	mvar group1_  Stream1_ 
	if ![info exists group1_] {
		puts "error: can't transmit before joining group!"
		exit 1
	} 
	if {$group1_ == $g} {
		$Stream1_ start 
	}
	mvar group2_ Stream2_
	if ![info exists group2_] {
		puts "error: can't transmit before joining group!"
		exit 1
	}
	if {$group2_ == $g} {
		$Stream2_ start 
	}
	mvar group3_ Stream3_
	if ![info exists group3_] {
		puts "error: can't transmit before joining group!"
		exit 1
	}
	if {$group3_ == $g} {
		$Stream3_ start 
	}
}


SimStream/RTP instproc stop {} {
	$self instvar Stream1_ Stream2_  Stream3_
	$Stream1_ stop
	$Stream2_ stop
	$Stream3_ stop
	$self set stopped_ 1
}

SimStream/RTP instproc stop-rcv {} {
	
	$self print-statistics
}

SimStream/RTP instproc stop-group { g } {
	mvar group1_  Stream1_ 
	if {$group1_ == $g} {
		$Stream1_ stop 
	}
	mvar group2_ Stream2_
	if {$group2_ == $g} {
		$Stream2_ stop 
	}
	mvar group3_ Stream3_
	if {$group3_ == $g} {
		$Stream3_ stop 
	}
}


SimStream/RTP instproc bye {} {
	mvar Stream1_ Stream2_ Stream3_ 
	$Stream1_ bye
	$Stream2_ bye
	$Stream3_ bye
}

SimStream/RTP instproc bye-group { g } {
	mvar group1_  Stream1_ 
	if {$group1_ == $g} {
		$Stream1_ bye 
	}
	mvar group2_ Stream2_
	if {$group2_ == $g} {
		$Stream2_ bye 
	}
	mvar group3_ Stream3_
	if {$group3_ == $g} {
		$Stream3_ bye 
	}
}

SimStream/RTP instproc go-up { l } {
	mvar group1_ group2_ group3_ Stream1_ Stream2_  Stream3_ level_
	if { $l == 1 } {
		
		$self bye-group $group1_ 
		$self leave-group $group1_
		$self set level_ 2 
		$self join-group $group2_
		$self start-session $group2_ 
		
		
	}
	if { $l == 2 } {
		
		$self bye-group $group2_ 
		$self leave-group $group2_
		$self set level_ 3 
		$self join-group $group3_
		$self start-session $group3_ 
		
		
	}
}

SimStream/RTP instproc go-down { l } {
	mvar group1_ group2_ group3_ Stream1_ Stream2_  Stream3_ level_
	if { $l == 3 } {
		
		$self bye-group $group3_ 
		$self leave-group $group3_
		$self set level_ 2
		$self join-group $group2_
		$self start-session $group2_
		
		
	}
	if { $l == 2 } {
		
		$self bye-group $group2_ 
		$self leave-group $group2_
		$self set level_ 1 
		$self join-group $group1_
		$self start-session $group1_ 
		
		
	}
}

SimStream/RTP instproc attach-node { node } {
	mvar Stream1_ Stream2_ Stream3_ 
	$Stream1_ attach-node $node 
	$Stream2_ attach-node $node 
	$Stream3_ attach-node $node 
	$self set node_ $node
}

SimStream/RTP instproc detach-node { node } {
	mvar Stream1_ Stream2_ Stream3_
	$Stream1_ detach-node $node 
	$Stream2_ detach-node $node
	$Stream3_ detach-node $node  
	$self unset node_
}

SimStream/RTP instproc join-groups { g1 g2 g3 } {
#	set g [expr $g]
	mvar group1_ group2_ group3_ Stream1_ Stream2_  Stream3_
	$self set group1_ $g1
	$Stream1_ join-group $g1
#	incr g
#	incr g
#	incr g
	$self set group2_ $g2
	$Stream2_ join-group $g2

	$self set group3_ $g3
	$Stream3_ join-group $g3
#	incr g
#	incr g
#	incr g
#	puts "Server <-----------------------------------"
#	puts [$Stream1_ set srcid_]
#	puts [$self set group1_]
#	puts [$Stream2_ set srcid_]
#	puts [$self set group2_]
#	puts "Server ----------------------------------->"
}

SimStream/RTP instproc join-group { g } {
	mvar group1_ group2_ group3_ Stream1_ Stream2_  Stream3_
#	set g [expr $g]
	if {$group1_ == $g} {
		$self set group1_ $g
		$Stream1_ join-group $g
	}
	mvar group2_ Stream2_
	if {$group2_ == $g} {
		$self set group2_ $g
		$Stream2_ join-group $g
	}
	mvar group3_ Stream3_
	if {$group3_ == $g} {
		$self set group3_ $g
		$Stream3_ join-group $g
	}
}


SimStream/RTP instproc leave-groups { } {
	mvar Stream1_ Stream2_ Stream3_ group1_ group2_ group3_
	$Stream1_ leave-group 
	$Stream2_ leave-group
	$Stream3_ leave-group  
	$self unset group1_
	$self unset group2_
	$self unset group3_
}

SimStream/RTP instproc leave-group { g } {
	mvar group1_  Stream1_ 
	if {$group1_ == $g} {
		$Stream1_ leave-group 
#		$self unset group1_
	}
	mvar group2_ Stream2_
	if {$group2_ == $g} {
		$Stream2_ leave-group 
#		$self unset group2_
	}
	mvar group3_ Stream3_
	if {$group3_ == $g} {
		$Stream3_ leave-group 
#		$self unset group3_
	}
}


SimStream/RTP instproc session_bw-low {l} {
	mvar Stream1_ 
	$Stream1_ session_bw $l 
}
	
SimStream/RTP instproc session_bw-medium {l} {
	mvar Stream2_ 
	$Stream2_ session_bw $l 
}

SimStream/RTP instproc session_bw-high {l} {
	mvar Stream3_ 
	$Stream3_ session_bw $l 
}

SimStream/RTP instproc transmit-low {l} {
	mvar Stream1_ 
	$Stream1_ transmit $l 
}
	
SimStream/RTP instproc transmit-medium {l} {
	mvar Stream2_ 
	$Stream2_ transmit $l 
}
	
SimStream/RTP instproc transmit-high {l} {
	mvar Stream3_ 
	$Stream3_ transmit $l 
}

SimStream/RTP instproc initial-rate-low {l} {
	mvar Stream1_ 
	$Stream1_ initial-rate $l 
}

SimStream/RTP instproc initial-rate-medium {l} {
	mvar Stream2_ 
	$Stream2_ initial-rate $l 
}

SimStream/RTP instproc initial-rate-high {l} {
	mvar Stream3_ 
	$Stream3_ initial-rate $l 
}

SimStream/RTP instproc enable-control {l} {
	mvar Stream1_  Stream2_ Stream3_
	$Stream1_ enable-control $l 
	$Stream2_ enable-control $l 
	$Stream3_ enable-control $l 
}

SimStream/RTP instproc low-stream-limits {low high} {
	mvar Stream1_ low_tx_ low_threshold_
	set l [bw_parse $low]
	set h [bw_parse $high]
	set mean [expr ($l)/8] 
	$self set low_tx_ $mean
	$self set low_threshold_ [expr $l/8]
#	$self set low_tx_ $l
	puts "$low_threshold_"
	$Stream1_ stream-limits $low $high 
	$Stream1_ set level_ 1
}

SimStream/RTP instproc medium-stream-limits {low high} {
	mvar Stream2_  medium_tx_ medium_threshold_
	set l [bw_parse $low]
	set h [bw_parse $high]
	set mean [expr ($l)/8] 
	$self set medium_tx_ $mean
	$self set medium_threshold_ [expr $l/8]
#	$self set medium_tx_ $l
	puts "$medium_threshold_"
	$Stream2_ stream-limits $low $high 
	$Stream2_ set level_ 2
}

SimStream/RTP instproc high-stream-limits {low high} {
	mvar Stream3_ high_tx_ high_threshold_
	set l [bw_parse $low]
	set h [bw_parse $high]
	set mean [expr ($l)/8] 
	$self set high_tx_ $mean
	$self set high_threshold_ [expr $l/8]
#	$self set high_tx_ $l
	puts "$high_threshold_"
	$Stream3_ stream-limits $low $high 
	$Stream3_ set level_ 3
}

#SimStream/RTP instproc start-feedback { } {
#	$self start-timer
#}

SimStream/RTP instproc streams-threshold { low medium high } {
	mvar low_threshold_  medium_threshold_ high_threshold_
	set l [bw_parse $low]
	set m [bw_parse $medium]
	set h [bw_parse $high]
	$self set low_threshold_ [expr $l/8]
	$self set medium_threshold_ [expr $m/8]
	$self set high_threshold_ [expr $h/8]
}
	
SimStream/RTP  set srcid_  -1



