#Author Georgios Kioumourtzis
# gkioumou@ceid.upatras.gr
# For Fig 2 on EUNICE 2009 paper
# 8 files will be created to make a plot
# assp0.tr : throughput of ASSP 0
# assp1.tr : throughput of ASSP 1
# assp2.tr : throughput of ASSP 2
# assp3.tr : throughput of ASSP 3
# assp4.tr : throughput of ASSP 4
# assp5.tr : throughput of ASSP 5
# tcp0.tr : throughput of TCP 1 
# tcp1.tr : throughput of TCP 2
# tcp2.tr : throughput of TCP 3
# throughput.tr : throughput of all receivers
# Delay and jitter measurements are also included


set ns [new Simulator -multicast on]


#set namfile [open out.nam w]
#$ns namtrace-all $namfile


set interval 1.0
set duration 250.0
set psize 1000
set fid 0

set tcp_num 0
set assp_recv_num 0
set num_assp_sink 6

set opt(low)	500kb
set opt(medium)	1000kb
set opt(high)   3000kb
set opt(report_interval) 200ms

#ns-random 0

set bw 1000
set num_links 4


Agent/TCP/Reno set packetSize_ $psize
Agent/TCP/Reno set overhead_ 0.002
#Agent/TCP set packetSize_ $psize
Agent/CBR/RTP  set packetSize_ $psize



proc setupTCP {start stop s r} {
    global ns tcp_num fid interval tcp tcp_sink ftp
    set tcp($tcp_num) [new Agent/TCP/Reno]
    $tcp($tcp_num) set fid_ $fid
    $tcp($tcp_num) set window_ 250
    set tcp_sink($tcp_num) [new Agent/TCPSink]
    $ns attach-agent $s $tcp($tcp_num)
    $ns attach-agent $r $tcp_sink($tcp_num)
    $ns connect $tcp($tcp_num) $tcp_sink($tcp_num)
    set ftp($tcp_num) [new Application/FTP]
    $ftp($tcp_num) attach-agent $tcp($tcp_num)
    $ns at $start "$ftp($tcp_num) start"
    incr tcp_num
    incr fid
    return $tcp([expr $tcp_num-1])
}


proc setupASSPSink { t l low medium high } {
	global ns assp_recv_num assp_sink duration
	set assp_sink($assp_recv_num) [new SimStream/RTP ]
	$assp_sink($assp_recv_num) set isReceiver true
	$assp_sink($assp_recv_num) set level_   $l 
	$assp_sink($assp_recv_num) report-interval 200ms
	$assp_sink($assp_recv_num) session_bw-low    $low
	$assp_sink($assp_recv_num) session_bw-medium $medium
	$assp_sink($assp_recv_num) session_bw-high   $high
	$assp_sink($assp_recv_num) initial-rate-low  10kb
	$assp_sink($assp_recv_num) initial-rate-medium  30kb
	$assp_sink($assp_recv_num) initial-rate-high  50kb
	$assp_sink($assp_recv_num) streams-threshold 1000kb 1800kb 3800kb
	$assp_sink($assp_recv_num) attach-node $t
	$ns at [expr $duration + 1] "$assp_sink($assp_recv_num) stop-rcv" 
	incr assp_recv_num
	return $assp_sink([expr $assp_recv_num-1])
}

set num_tcp 3
set n2 [$ns node]
set n3 [$ns node]
set n4 [$ns node]
set n5 [$ns node]
set n6 [$ns node]
for {set i 0} {$i < $num_tcp} {incr i} {
	set tcp_s($i) [$ns node]
	set tcp_r($i) [$ns node]
	$ns duplex-link $tcp_s($i) $n2 100Mb 8ms DropTail
}
set assp_s [$ns node]
$assp_s label "ASSP Sender"

for {set i 0} {$i < 6 } {incr i} {
set assp_r($i) [$ns node]
$assp_r($i) label "ASSP Receiver_($i)"
}


# Use automatic layout
$ns duplex-link $assp_s $n2 100Mb 8ms DropTail
$ns duplex-link $n2 $n3 100Mb 8ms DropTail
$ns duplex-link $n3 $n4 10Mb 8ms DropTail
$ns duplex-link $n3 $n5 5Mb 8ms DropTail
$ns duplex-link $n3 $n6 2Mb 8ms DropTail
#$ns queue-limit $n3 $n5 42 
#$ns queue-parameter $n3 $n5 2 21
$ns duplex-link $n4 $assp_r(0) 10Mb 8ms DropTail
$ns duplex-link $n4 $assp_r(1) 10Mb 8ms DropTail
$ns duplex-link $n4 $tcp_r(0) 10Mb 8ms DropTail

$ns duplex-link $n5 $assp_r(2) 10Mb 8ms DropTail
$ns duplex-link $n5 $assp_r(3) 10Mb 8ms DropTail
$ns duplex-link $n5 $tcp_r(1)  10Mb 8ms DropTail

$ns duplex-link $n6 $assp_r(4) 10Mb 8ms DropTail
$ns duplex-link $n6 $assp_r(5) 10Mb 8ms DropTail
$ns duplex-link $n6 $tcp_r(2)  10Mb 8ms DropTail



#set queue_size [expr 40*4/8]
DM set PruneTimeout 600.0
set mproto DM
set mrthandle [$ns mrtproto $mproto]
set group1 [Node allocaddr]
set group2 [Node allocaddr]
set group3 [Node allocaddr]
#set mrthandle [$ns mrtproto $mproto {}]
#$ns queue-limit $n2 $n3 $queue_size

set s0 [new SimStream/RTP ]
$s0 set isSender true
#$s0 report-interval $opt(report_interval)
$s0 session_bw-low    $opt(low)
$s0 session_bw-medium    $opt(medium)
$s0 session_bw-high    $opt(high)
$s0 initial-rate-low  100kb
$s0 initial-rate-medium  500kb
$s0 initial-rate-high  1000kb

$s0 low-stream-limits 		100kb 2000kb
$s0 medium-stream-limits 	2000kb 4000kb
$s0 high-stream-limits 		4Mb 8Mb
$s0 attach-node $assp_s
# enable congestion control
$s0 enable-control 1


setupASSPSink $assp_r(0) 1 1Mb 2Mb 4Mb
setupASSPSink $assp_r(1) 1 1Mb 2Mb 4Mb
setupASSPSink $assp_r(2) 1 1Mb 2Mb 4Mb
setupASSPSink $assp_r(3) 1 1Mb 2Mb 4Mb
setupASSPSink $assp_r(4) 1 1Mb 2Mb 4Mb
setupASSPSink $assp_r(5) 1 1Mb 2Mb 4Mb

$ns at 0.0 "$s0 join-groups $group1 $group2 $group3"
$ns at 0.0 "$s0 start"





$ns at 0.0 "$assp_sink(0) assign-groups $group1 $group2 $group3"
$ns at 0.0 "$assp_sink(0) join-group $group1"
$ns at 0.0 "$assp_sink(0) start-session $group1"


$ns at 0.0 "$assp_sink(1) assign-groups $group1 $group2 $group3"
$ns at 0.0 "$assp_sink(1) join-group $group1"
$ns at 0.0 "$assp_sink(1) start-session $group1"


$ns at 0.4 "$assp_sink(2) assign-groups $group1 $group2 $group3"
$ns at 0.4 "$assp_sink(2) join-group $group1"
$ns at 0.4 "$assp_sink(2) start-session $group1"


$ns at 0.0 "$assp_sink(3) assign-groups $group1 $group2 $group3"
$ns at 0.0 "$assp_sink(3) join-group $group1"
$ns at 0.0 "$assp_sink(3) start-session $group1"


$ns at 0.0 "$assp_sink(4) assign-groups $group1 $group2 $group3"
$ns at 0.0 "$assp_sink(4) join-group $group1"
$ns at 0.0 "$assp_sink(4) start-session $group1"


$ns at 0.0 "$assp_sink(5) assign-groups $group1 $group2 $group3"
$ns at 0.0 "$assp_sink(5) join-group $group1"
$ns at 0.0 "$assp_sink(5) start-session $group1"


$ns at 0.0 "$s0 transmit-low		$opt(low)"
$ns at 0.0 "$s0 transmit-medium		$opt(medium)"
$ns at 0.0 "$s0 transmit-high		$opt(high)"




for {set i 0} {$i < $num_tcp} {incr i} {
    setupTCP 0.0 $duration $tcp_s($i) $tcp_r($i)
}

for {set i 0} {$i < $num_tcp} {incr i} {
          set f_tcp($i)  [open tcp$i.tr w]
}

set a_tcp  [open avg_tcp.tr w]

for {set i 0} {$i < $num_assp_sink} {incr i} {
          set f_assp($i)  [open assp$i.tr w]
}
for {set i 0} {$i < $num_assp_sink} {incr i} {
          set delay_assp($i)  [open delay_assp$i.tr w]
}

for {set i 0} {$i < $num_assp_sink} {incr i} {
          set jitter_assp($i)  [open jitter_assp$i.tr w]
}

proc record {} {

        global n ns  tcp_sink f_tcp num_tcp assp_sink f_assp a_tcp delay_assp jitter_assp
          #Get an instance of the simulator
          set ns [Simulator instance]
          #Set the time after which the procedure should be called again
          set time 0.5
          #How many bytes have been received by the traffic sinks?
          #Get the current time
          set now [$ns now]
        for {set i 0} {$i < $num_tcp } {incr i} {
          set bw_tcp($i) [$tcp_sink($i) set bytes_]

          #Calculate the bandwidth (in MBit/s) and write it to the files
          set bd_tcp($i)  [expr $bw_tcp($i)/$time*8/1000000]
          puts $f_tcp($i) "$now  $bd_tcp($i)"

          #Reset the bytes_ values on the traffic sinks
          $tcp_sink($i) set bytes_ 0
        }
          set avg_bd_tcp 0 
          set n_tcp [expr $num_tcp -1] 

        for {set i 2} {$i < $num_tcp } {incr i} {
          set avg_bd_tcp  [expr $bd_tcp($i)+$avg_bd_tcp]
	}
          set avg_bd_tcp  [expr $avg_bd_tcp/$n_tcp]
          puts $a_tcp "$now  $avg_bd_tcp"

        for {set i 0} {$i < 6 } {incr i} {
	  set bw_assp($i) [$assp_sink($i) set rcv_bytes_]
          set rate_assp($i) [expr $bw_assp($i)/$time*8/1000000]
          puts $f_assp($i) "$now $rate_assp($i)"
	  $assp_sink($i) set rcv_bytes_ 0
         
        }
	 for {set i 0} {$i < 6 } {incr i} {
          set  delay_($i) [$assp_sink($i) set end_2_end_delay_]
          puts $delay_assp($i) "$now $delay_($i)"
         
        }
	for {set i 0} {$i < 6 } {incr i} {
          set  jitter_($i) [$assp_sink($i) set jitter_]
	  puts $jitter_assp($i) "$now $jitter_($i)"
         
        }

          #Re-schedule the procedure
          $ns at [expr $now+$time] "record"
  }

proc finish {} {
        global ns duration num_assp_sink assp_sink f_assp delay_assp jitter_assp f_tcp
        global tcp_sink  num_tcp  ;#namfile
        global start_measure
        set f_throughput  [open throughput.tr w]
       # set f_avg_throughput  [open avg.tr w]

	 for {set i 0} {$i < $num_assp_sink} {incr i} {
                set tot_assp($i) [$assp_sink($i) set total_pkts_recv_]
                set S_Th_b($i) [expr ($tot_assp($i)*1000)/($duration)*8/1000000]
                puts "ASSP($i)  Throughput  $S_Th_b($i)"
                puts $f_throughput "ASSP $i $S_Th_b($i)"
               
        }
	 for {set i 0} {$i < $num_assp_sink} {incr i} {
		close $f_assp($i)
		close $delay_assp($i)
		close $jitter_assp($i)
	}
	for {set i 0} {$i < 3} {incr i} {
		close $f_tcp($i)
	}

        for {set i 0} {$i < 3} {incr i} {
                set tot_tcp($i) [$tcp_sink($i) set total_bytes_]
                set Tcp_Th_b($i) [expr $tot_tcp($i)/($duration - $start_measure)*8/1000000]
                puts "Tcp($i) Throughput after 20s $Tcp_Th_b($i)"
                puts $f_throughput "TCP $i $Tcp_Th_b($i)"
        }
	
	 	
        
	
	#$ns flush-trace
     	#close $namfile
	exec xgraph    tcp0.tr tcp1.tr tcp2.tr -geometry 800x400  &
	exec xgraph    assp0.tr assp1.tr assp2.tr assp3.tr assp4.tr assp5.tr  -geometry 800x400  &
	exec xgraph    assp0.tr assp1.tr  -geometry 800x400  &
	exec xgraph    assp2.tr assp3.tr  -geometry 800x400  &
	exec xgraph    assp4.tr assp5.tr  -geometry 800x400  &
	exit 0
}

proc start_measurement {} {
        global ns num_tcp  tcp_sink
        
        for {set i 0} {$i < $num_tcp} {incr i} {
          $tcp_sink($i) set total_bytes_ 0
        }
}

set start_measure 10 
$ns at 0.0 "record"
$ns at [expr $duration + 2] "finish"
$ns at $start_measure  "start_measurement"
$ns run
