#Author Georgios Kioumourtzis
# gkioumou@ceid.upatras.gr
# For Fig 3 on EUNICE 2009 paper
# 5 files will be created to make a plot
# smcc0.tr : throughput of SMCC 1
# smcc1.tr : throughput of SMCC 2
# smcc2.tr : throughput of SMCC 3
# smcc3.tr : throughput of SMCC 4
# smcc4.tr : throughput of SMCC 5
# smcc5.tr : throughput of SMCC 6
# tcp0.tr : throughput of TCP 1 
# tcp1.tr : throughput of TCP 2
# tcp2.tr : throughput of TCP 3 
# throughput.tr : throughput of all receivers

set ns [new Simulator -multicast on]


#set namfile [open out.nam w]
#$ns namtrace-all $namfile

#Open the NS trace file
#set tracefile [open out.tr w]
#$ns trace-all $tracefile

set interval 1.0
set duration 250.0
set psize 1000
set fid 0

set tcp_num 0
set smcc_num 0
set smcc_recv_num 0

#ns-random 0

set bw 1000
set num_links 4


Agent/TCP set packetSize_ $psize
Agent/SMCC set packetSize_ $psize
Agent/TCP set overhead_ 0.002
Agent/SMCC set base_layer_rate 125000 
Agent/SMCCSink set base_layer_rate 125000

proc setupTCP {start stop s r} {
    global ns tcp_num fid interval tcp tcp_sink ftp
    set tcp($tcp_num) [new Agent/TCP/Reno]
    $tcp($tcp_num) set fid_ $fid
    $tcp($tcp_num) set window_ 250
    set tcp_sink($tcp_num) [new Agent/TCPSink]
    $ns attach-agent $s $tcp($tcp_num)
    $ns attach-agent $r $tcp_sink($tcp_num)
    $ns connect $tcp($tcp_num) $tcp_sink($tcp_num)
    set ftp($tcp_num) [new Application/FTP]
    $ftp($tcp_num) attach-agent $tcp($tcp_num)
    $ns at $start "$ftp($tcp_num) start"
    incr tcp_num
    incr fid
    return $tcp([expr $tcp_num-1])
}

proc setupSMCC {start stop s} {
    global ns smcc_num fid interval smcc
    set start2 [expr $start + 1]
    set smcc($smcc_num) [new Agent/SMCC]
    $smcc($smcc_num) set printStatus_  1
    $smcc($smcc_num) set fid_ $fid
    $smcc($smcc_num) set dst_port_ 0
    $ns attach-agent $s $smcc($smcc_num)
    $ns at $start "$smcc($smcc_num) warmup"
    $ns at $start2 "$smcc($smcc_num) start"
   
    incr smcc_num
    incr fid
    return $smcc([expr $smcc_num-1])
}

proc setupSMCCSink {start stop t r} {
    global ns smcc_recv_num smcc_sink
    set smcc_sink($smcc_recv_num) [new Agent/SMCCSink]
    $smcc_sink($smcc_recv_num) set printStatus_  1
    $ns attach-agent $r $smcc_sink($smcc_recv_num)
    $smcc_sink($smcc_recv_num) set dst_addr_ [$t set agent_addr_]
    $smcc_sink($smcc_recv_num) set dst_port_ [$t set agent_port_]
    $smcc_sink($smcc_recv_num) set recv_id_ $smcc_recv_num
    $ns at $start "$smcc_sink($smcc_recv_num) begin"
    $ns at $stop  "$smcc_sink($smcc_recv_num) stop"
    incr smcc_recv_num
    return $smcc_sink([expr $smcc_recv_num-1])
}



#set namfile [open out.nam w]
#$ns namtrace-all $namfile


set num_tcp 3
set n2 [$ns node]
set n3 [$ns node]
set n4 [$ns node]
set n5 [$ns node]
set n6 [$ns node]
for {set i 0} {$i < $num_tcp} {incr i} {
	set tcp_s($i) [$ns node]
	set tcp_r($i) [$ns node]
	$ns duplex-link $tcp_s($i) $n2 100Mb 8ms DropTail
}
set smcc_s [$ns node]
$smcc_s label "SMCC Sender"
for {set i 0} {$i < 6 } {incr i} {
set smcc_r($i) [$ns node]
$smcc_r($i) label "SMCC Receiver"
}


# Use automatic layout
$ns duplex-link $smcc_s $n2 100Mb 8ms DropTail
$ns duplex-link $n2 $n3 100Mb 8ms DropTail
$ns duplex-link $n3 $n4 10Mb 8ms DropTail
$ns duplex-link $n3 $n5 5Mb 8ms DropTail
$ns duplex-link $n3 $n6 2Mb 8ms DropTail
#$ns queue-limit $n3 $n5 42 
#$ns queue-parameter $n3 $n5 2 21
$ns duplex-link $n4 $smcc_r(0) 10Mb 8ms DropTail
$ns duplex-link $n4 $smcc_r(1) 10Mb 8ms DropTail
$ns duplex-link $n4 $tcp_r(0) 10Mb 8ms DropTail

$ns duplex-link $n5 $smcc_r(2) 10Mb 8ms DropTail
$ns duplex-link $n5 $smcc_r(3) 10Mb 8ms DropTail
$ns duplex-link $n5 $tcp_r(1)  10Mb 8ms DropTail

$ns duplex-link $n6 $smcc_r(4) 10Mb 8ms DropTail
$ns duplex-link $n6 $smcc_r(5) 10Mb 8ms DropTail
$ns duplex-link $n6 $tcp_r(2)  10Mb 8ms DropTail



#set queue_size [expr 40*4/8]
DM set PruneTimeout 600.0
set mproto DM
set mrthandle [$ns mrtproto $mproto]
#set mrthandle [$ns mrtproto $mproto {}]
#$ns queue-limit $n2 $n3 $queue_size

set t [setupSMCC 1.0 $duration $smcc_s]
set num_smcc_sink 6
setupSMCCSink 3.5 $duration $t $smcc_r(0)
setupSMCCSink 3.5 $duration $t $smcc_r(1)
setupSMCCSink 3.5 $duration $t $smcc_r(2)
setupSMCCSink 3.5 $duration $t $smcc_r(3)
setupSMCCSink 3.5 $duration $t $smcc_r(4)
setupSMCCSink 3.5 $duration $t $smcc_r(5)


for {set i 0} {$i < $num_tcp} {incr i} {
    setupTCP 5.0 $duration $tcp_s($i) $tcp_r($i)
}

for {set i 0} {$i < $num_tcp} {incr i} {
          set f_tcp($i)  [open tcp$i.tr w]
}

set a_tcp  [open avg_tcp.tr w]

for {set i 0} {$i < $num_smcc_sink} {incr i} {
          set f_smcc($i)  [open smcc$i.tr w]
}

for {set i 0} {$i < $num_smcc_sink} {incr i} {
          set delay_smcc($i)  [open delay_smcc$i.tr w]
}
for {set i 0} {$i < $num_smcc_sink} {incr i} {
          set jitter_smcc($i)  [open jitter_smcc$i.tr w]
}

proc record {} {

        global n ns  tcp_sink f_tcp num_tcp num_smcc_sink  smcc_sink f_smcc a_tcp delay_smcc jitter_smcc
          #Get an instance of the simulator
          set ns [Simulator instance]
          #Set the time after which the procedure should be called again
          set time 0.5
          #How many bytes have been received by the traffic sinks?
          #Get the current time
          set now [$ns now]
        for {set i 0} {$i < $num_tcp } {incr i} {
          set bw_tcp($i) [$tcp_sink($i) set bytes_]

          #Calculate the bandwidth (in MBit/s) and write it to the files
          set bd_tcp($i)  [expr $bw_tcp($i)/$time*8/1000000]
          puts $f_tcp($i) "$now  $bd_tcp($i)"

          #Reset the bytes_ values on the traffic sinks
          $tcp_sink($i) set bytes_ 0
        }
          set avg_bd_tcp 0 
          set n_tcp [expr $num_tcp -1] 

        for {set i 2} {$i < $num_tcp } {incr i} {
          set avg_bd_tcp  [expr $bd_tcp($i)+$avg_bd_tcp]
	}
          set avg_bd_tcp  [expr $avg_bd_tcp/$n_tcp]
          puts $a_tcp "$now  $avg_bd_tcp"

        for {set i 0} {$i < 6 } {incr i} {
          set rate_smcc($i) [$smcc_sink($i) set display_rate_]
          puts $f_smcc($i) "$now [expr $rate_smcc($i)/$time*8/1000000]"
          $smcc_sink($i) set display_rate_ 0
        }
	
	 for {set i 0} {$i < 6 } {incr i} {
          set  delay_($i) [$smcc_sink($i) set end_2_end_delay_]
          puts $delay_smcc($i) "$now $delay_($i)"
         
        }
	for {set i 0} {$i < 6 } {incr i} {
          set  jitter_($i) [$smcc_sink($i) set jitter_]
          puts $jitter_smcc($i) "$now $jitter_($i)"
         
        }
	
	#Re-schedule the procedure
          $ns at [expr $now+$time] "record"
  }

proc finish {} {
        global ns duration 
        global num_smcc_sink tcp_sink  num_tcp  smcc_sink f_smcc delay_smcc jitter_smcc f_tcp;#tracefile namfile
        global start_measure
        set f_throughput  [open throughput.tr w]
        set f_avg_throughput  [open avg.tr w]

        for {set i 0} {$i < $num_smcc_sink} {incr i} {
                set tot_smcc($i) [$smcc_sink($i) set total_bytes_]
                set S_Th_b($i) [expr $tot_smcc($i)/($duration - $start_measure)*8/1000000]
                puts "SMCC($i)  Throughput  after 20s $S_Th_b($i)"
                puts $f_throughput "SMCC $i $S_Th_b($i)"
                puts $f_avg_throughput "SMCC $i $S_Th_b($i)"
        }
	 for {set i 0} {$i < $num_smcc_sink} {incr i} {
		close $f_smcc($i)
		close $delay_smcc($i)
		close $jitter_smcc($i)
	}
	for {set i 0} {$i < 3} {incr i} {
		close $f_tcp($i)
	}
        for {set i 0} {$i < 3} {incr i} {
                set tot_tcp($i) [$tcp_sink($i) set total_bytes_]
                set Tcp_Th_b($i) [expr $tot_tcp($i)/($duration - $start_measure)*8/1000000]
                puts "Tcp($i) Throughput after 20s $Tcp_Th_b($i)"
                puts $f_throughput "TCP $i $Tcp_Th_b($i)"
        }
	#$ns flush-trace
     	#close $tracefile
	#close $namfile
	exec xgraph    tcp0.tr tcp1.tr tcp2.tr -geometry 800x400  &
	exec xgraph    smcc0.tr smcc3.tr smcc5.tr  -geometry 800x400  &
	exit 0
}

proc start_measurement {} {
        global n ns  num_smcc_sink  num_tcp smcc_sink tcp_sink
        for {set i 0} {$i < $num_smcc_sink} {incr i} {
          $smcc_sink($i) set total_bytes_ 0
        }
        for {set i 0} {$i < $num_tcp} {incr i} {
          $tcp_sink($i) set total_bytes_ 0
        }
}



set start_measure 10 
$ns at 0.0 "record"
$ns at $duration "finish"
$ns at $start_measure  "start_measurement"
$ns run
